"""
Selector tests.
"""

from unittest.mock import Mock
import pytest
from lute.ankiexport.criteria import evaluate_criteria, validate_criteria
from lute.ankiexport.exceptions import AnkiExportConfigurationError


@pytest.fixture(name="term")
def fixture_term():
    "Fake term."
    parent = Mock(text="PARENT")
    parent.term_tags = [Mock(text="parenttag")]

    term = Mock()
    term.id = 1
    term.text = "HOLA"
    term.status = 3
    term.language.name = "Spanish"
    term.language.id = 42
    term.get_current_image.return_value = "image.jpg"
    term.parents = [parent]
    term.term_tags = [Mock(text="masc"), Mock(text="xxx")]
    term.translation = "example translation"
    return term


@pytest.mark.parametrize(
    "criteria,expected",
    [
        ('language:"Spanish"', True),
        ('language:"xxx"', False),
        ("parents.count=1", True),
        ("parents.count==1", True),
        ("parents.count>=0", True),
        ("parents.count>1", False),
        ('tags:"masc"', True),
        ('tags:"fem"', False),
        ('tags:["fem", "masc"]', True),
        ("status<=3", True),
        ("status==1", False),
        ('tags:["fem", "other"]', False),
        ('tags:["parenttag"]', False),
        ('parents.tags:["parenttag"]', True),
        ('all.tags:["parenttag"]', True),
        ('parents.count=1 and tags:["fem", "other"] and status<=3', False),
    ],
)
def test_criteria(criteria, expected, term):
    "Check criteria vs test term."
    assert evaluate_criteria(criteria, term) == expected, criteria


def test_blank_criteria_is_always_true(term):
    assert evaluate_criteria("", term) is True, "blank"
    assert evaluate_criteria(None, term) is True, "None"


@pytest.mark.parametrize(
    "criteria",
    [
        ('lanxguage:"Spanish"'),
        ('language="xxx"'),
        ("parents=1"),
        ('tags="masc"'),
        ('tags["fem", "masc"]'),
        ('parents.count=1 and tags["fem", "other"] and status<=3'),
    ],
)
def test_bad_criteria_throws(criteria, term):
    "Check criteria vs test term."
    with pytest.raises(AnkiExportConfigurationError):
        evaluate_criteria(criteria, term)


@pytest.mark.parametrize(
    "criteria",
    [
        ('lanxguage:"Spanish"'),
        ('language="xxx"'),
        ("parents=1"),
        ('tags="masc"'),
        ('tags["fem", "masc"]'),
        ('parents.count=1 and tags["fem", "other"] and status<=3'),
    ],
)
def test_validate_criteria_throws_if_bad(criteria):
    "Check criteria vs test term."
    with pytest.raises(AnkiExportConfigurationError):
        validate_criteria(criteria)



"""
Language dictionaries markdown table generation.
"""

from glob import glob
import os
import yaml


def get_table_rows():
    "Get markdown table for docs."
    thisdir = os.path.dirname(__file__)
    langglob = os.path.join(thisdir, "**", "definition.yaml")
    output = []
    for filename in glob(langglob):
        with open(filename, "r", encoding="utf-8") as f:
            y = yaml.safe_load(f)
            langname = y["name"]
            for d in y["dictionaries"]:
                data = [langname, d["for"], d["type"], f"`{d['url']}`"]
                output.append("|" + "|".join(data) + "|")
    output.sort()
    return output


def generate_table():
    "Gen table"
    print("|Language|For|Type|URL|")
    print("|---|---|---|---|")
    rows = get_table_rows()
    print("\n".join(rows))


if __name__ == "__main__":
    generate_table()



"""
Term Repository tests.

Tests lute.term.model.Term *domain* objects being saved
and retrieved from DB.
"""

import pytest

from lute.models.term import Term as DBTerm, TermTag
from lute.db import db
from lute.term.model import Term, Repository
from tests.dbasserts import assert_sql_result, assert_record_count_equals
from tests.utils import add_terms


@pytest.fixture(name="repo")
def fixture_repo():
    return Repository(db.session)


@pytest.fixture(name="hello_term")
def fixture_hello_term(english):
    """
    Term business object with some defaults,
    no tags or parents.
    """
    t = Term()
    t.language_id = english.id
    t.text = "HELLO"
    t.translation = "greeting"
    t.current_image = "hello.png"
    t.flash_message = "hello flash"
    return t


def test_save_new(app_context, hello_term, repo):
    """
    Saving a simple Term object loads the database.
    """
    sql = "select WoText, WoTextLC, WoTokenCount from words"
    assert_sql_result(sql, [], "empty table")

    repo.add(hello_term)
    assert_sql_result(sql, [], "Still empty")

    repo.commit()
    assert_sql_result(sql, ["HELLO; hello; 1"], "Saved")

    term = repo.find(hello_term.language_id, hello_term.text)
    assert term.text == hello_term.text


def test_save_new_multiword(app_context, hello_term, repo):
    """
    Zero-width strings are added between parsed tokens.
    """
    sql = "select WoText, WoTextLC, WoTokenCount from words"
    assert_sql_result(sql, [], "empty table")

    hello_term.text = "HELLO THERE"
    repo.add(hello_term)
    assert_sql_result(sql, [], "Still empty")

    repo.commit()
    # Assert replaces zws with '/'
    assert_sql_result(sql, ["HELLO/ /THERE; hello/ /there; 3"], "Saved")


def test_save_updates_existing(english, app_context, hello_term, repo):
    """
    Saving Term updates an existing term if the text and lang
    matches.
    """

    term = DBTerm(english, "HELLO")
    db.session.add(term)
    db.session.commit()
    sql = "select WoID, WoText, WoStatus from words"
    assert_sql_result(sql, [f"{term.id}; HELLO; 1"], "have term")

    hello_term.language_id = english.id
    hello_term.text = "hello"
    hello_term.status = 5

    repo.add(hello_term)
    repo.commit()
    assert_sql_result(sql, [f"{term.id}; hello; 5"], "have term, status changed")


def test_save_existing_replaces_tags(english, app_context, repo):
    """
    If the db has a term with tags, and the business term changes
    those, the existing tags are replaced.
    """
    dbterm = DBTerm(english, "HELLO")
    dbterm.add_term_tag(TermTag("a"))
    dbterm.add_term_tag(TermTag("b"))
    db.session.add(dbterm)
    db.session.commit()
    sql = f"""select TgText from tags
    inner join wordtags on WtTgID = TgID
    where WtWoID = {dbterm.id}"""
    assert_sql_result(sql, ["a", "b"], "have term tags")

    term = repo.load(dbterm.id)
    assert term.term_tags == ["a", "b"]

    term.term_tags = ["a", "c"]
    repo.add(term)
    repo.commit()
    assert_sql_result(sql, ["a", "c"], "term tags changed")

    term.term_tags = []
    repo.add(term)
    repo.commit()
    assert_sql_result(sql, [], "term tags removed")

    sql = "select TgText from tags order by TgText"
    assert_sql_result(sql, ["a", "b", "c"], "Source tags still exist")


def test_save_uses_existing_TermTags(app_context, repo, hello_term):
    "Don't create new TermTag records if they already exist."
    db.session.add(TermTag("a"))
    db.session.commit()

    sql = """select TgID, TgText, WoText
    from tags
    left join wordtags on WtTgID = TgID
    left join words on WoID = WtWoID
    order by TgText"""
    assert_sql_result(sql, ["1; a; None"], "a tag exists")

    hello_term.term_tags = ["a", "b"]
    repo.add(hello_term)
    repo.commit()
    assert_sql_result(sql, ["1; a; HELLO", "2; b; HELLO"], "a used, b created")


def test_fix_issue_454_handle_duplicate_tags(app_context, repo, hello_term):
    "Same new tag added twice should be handled ok."
    db.session.add(TermTag("a"))
    db.session.commit()

    sql = """select TgID, TgText, WoText
    from tags
    left join wordtags on WtTgID = TgID
    left join words on WoID = WtWoID
    order by TgText"""
    assert_sql_result(sql, ["1; a; None"], "a tag exists")

    hello_term.term_tags = ["a", "b", "b"]
    repo.add(hello_term)
    repo.commit()
    assert_sql_result(sql, ["1; a; HELLO", "2; b; HELLO"], "a used, b created")


def test_save_with_no_flash_message(app_context, repo, hello_term):
    "Saving with flash = None removes the flash record."
    hello_term.flash_message = "hi there"
    repo.add(hello_term)
    repo.commit()
    sql = "select WfMessage from wordflashmessages"
    assert_sql_result(sql, ["hi there"], "record exists")

    hello_term.flash_message = None
    repo.add(hello_term)
    repo.commit()
    assert_sql_result(sql, [], "removed")


## Deletes.


def test_delete(app_context, repo, hello_term):
    "Removes record from db."
    repo.add(hello_term)
    repo.commit()
    sql = "select WoTextLC, WoStatus from words"
    assert_sql_result(sql, ["hello; 1"], "record exists")

    repo.delete(hello_term)
    repo.commit()
    assert_sql_result(sql, [], "deleted")


def test_delete_leaves_parent(app_context, repo, hello_term):
    "Parent is left."
    hello_term.parents.append("parent")
    repo.add(hello_term)
    repo.commit()
    sql = "select WoTextLC from words"
    assert_sql_result(sql, ["hello", "parent"], "records exists")

    repo.delete(hello_term)
    repo.commit()
    assert_sql_result(sql, ["parent"], "parent stays")


## Saving and images.


def test_save_term_image_saved_and_replaced(app_context, repo, hello_term):
    "Saving puts record in image table."
    sql = "select WiSource from wordimages"
    assert_sql_result(sql, [], "nothing present")
    repo.add(hello_term)
    repo.commit()
    assert_sql_result(sql, ["hello.png"], "image saved")


def test_save_term_image_set_to_blank_removes_record(app_context, repo, hello_term):
    "Saving puts record in image table."
    sql = "select WiSource from wordimages"
    repo.add(hello_term)
    repo.commit()
    assert_sql_result(sql, ["hello.png"], "image saved")

    hello_term.current_image = ""
    repo.add(hello_term)
    repo.commit()
    assert_sql_result(sql, [], "record removed")

    hello_term.current_image = "new.png"
    repo.add(hello_term)
    repo.commit()
    assert_sql_result(sql, ["new.png"], "image saved")

    hello_term.current_image = None
    repo.add(hello_term)
    repo.commit()
    assert_sql_result(sql, [], "record removed again")


## Saving and parents.


def create_parent(lang, status=0, translation=None, image=None, tags=None):
    "Create test parent."
    p = DBTerm(lang, "parent")
    p.status = status
    p.translation = translation
    p.set_current_image(image)
    for t in tags or []:
        p.add_term_tag(TermTag(t))  # Ensure no double-tag added.
    db.session.add(p)
    db.session.commit()
    return p


def assert_parent_has(status, translation, img, term_tags):
    "Assert parent data matches expected."
    # For some weird reason, query with filter was not returning
    # a match, so doing a loop to get the parent. Possibly due to aliasing
    # of lute.models.term.Term to DBTerm.
    # p = db.session.query(DBTerm).filter(DBTerm.text == "parent").first()
    parents = [t for t in db.session.query(DBTerm).all() if t.text == "parent"]
    assert len(parents) == 1, "Sanity check"
    p = parents[0]
    assert p.status == status, "status"
    assert (p.translation or "-") == (translation or "-"), "txn"
    assert (p.get_current_image() or "-") == (img or "-"), "img"
    actual_tags = sorted([t.text for t in p.term_tags])
    assert actual_tags == term_tags, "tags"


def test_save_new_child_creates_new_populated_parent(app_context, repo, hello_term):
    """
    Given a Term with parents = [ 'newparent' ],
    new parent DBTerm is created, and is assigned translation and image and tag.

    The child's data is propagated up if needed, to 'fill in' missing parent data.
    """
    t = hello_term
    t.parents = ["parent"]
    t.term_tags = ["a", "b"]
    repo.add(t)
    repo.commit()

    assert_parent_has(t.status, t.translation, t.current_image, ["a", "b"])


def test_save_new_child_populates_existing_unknown_parent(
    app_context, repo, english, hello_term
):
    "Existing parent with status 0 is bumped to status 1."
    create_parent(english, status=0)

    hello_term.parents = ["parent"]
    hello_term.term_tags = ["a", "b"]
    repo.add(hello_term)
    repo.commit()

    t = hello_term
    assert_parent_has(t.status, t.translation, t.current_image, ["a", "b"])


def test_save_new_child_sets_existing_parent_translation_and_image_if_missing(
    app_context, repo, english, hello_term
):
    "Existing new child data is propagated up if needed, to 'fill in' missing parent data."
    create_parent(english, status=3, translation="something")

    hello_term.parents = ["parent"]
    hello_term.term_tags = ["a", "b"]
    repo.add(hello_term)
    repo.commit()

    assert_parent_has(3, "something", hello_term.current_image, [])


def test_save_existing_child_creates_new_populated_parent(
    app_context, repo, hello_term
):
    """
    Given a Term with parents = [ 'newparent' ],
    new parent DBTerm is created, and is assigned translation and image and tag.

    The child's data is propagated up if needed, to 'fill in' missing parent data.
    """
    repo.add(hello_term)
    repo.commit()
    assert_sql_result("select WoText from words", ["HELLO"], "no parent yet")

    hello_term.parents = ["parent"]
    hello_term.term_tags = ["a", "b"]
    repo.add(hello_term)
    repo.commit()

    t = hello_term
    assert_parent_has(t.status, t.translation, t.current_image, ["a", "b"])


def test_save_existing_child_populates_existing_unknown_parent(
    app_context, repo, english, hello_term
):
    "Existing parent with status 0 is bumped to status 1."
    create_parent(english, status=0, tags=["a"])

    repo.add(hello_term)
    repo.commit()

    parent = repo.find(hello_term.language_id, "parent")
    assert parent.status == 0, "parent still unknown"

    h2 = repo.find(hello_term.language_id, hello_term.text)
    h2.parents = ["parent"]
    h2.term_tags = ["a", "b"]
    repo.add(h2)
    repo.commit()

    t = hello_term
    assert_parent_has(t.status, t.translation, t.current_image, ["a", "b"])


def test_update_child_add_existing_parent_does_not_change_parent_data_even_if_missing(
    app_context, repo, english, hello_term
):
    """
    If a parent existed before, and the child existed before, then
    editing the child shouldn't affect the parent, even if the
    parent's translation and image are empty -- reason: they have been
    created and are specifically empty.
    """
    create_parent(english, translation=None, image=None, status=3)

    repo.add(hello_term)
    repo.commit()

    hello_term = repo.find(hello_term.language_id, hello_term.text)
    assert hello_term is not None, "Have hello_term"
    hello_term.parents = ["parent"]
    hello_term.term_tags = ["a", "b"]
    repo.add(hello_term)
    repo.commit()

    assert_parent_has(3, None, None, [])


def test_update_child_with_parent_does_not_change_parent_data_even_if_missing(
    app_context, repo, english, hello_term
):
    """
    If a parent existed before, and the child existed before, then
    editing the child shouldn't affect the parent, even if the
    parent's translation and image are empty -- reason: they have been
    created and are specifically empty.
    """
    p = create_parent(english, translation=None, image=None, status=3)

    hello_term.parents = ["parent"]
    repo.add(hello_term)
    repo.commit()

    # Parent updated on initial save.
    assert_parent_has(3, hello_term.translation, hello_term.current_image, [])

    # Re-set parent.
    p.translation = None
    p.set_current_image(None)
    db.session.add(p)
    db.session.commit()
    assert_parent_has(3, None, None, [])

    # Re-update existing child term.
    hello_term = repo.find(hello_term.language_id, hello_term.text)
    assert hello_term.parents == ["parent"], "parent still set"
    hello_term.translation = "UPDATED"
    hello_term.current_image = "UPDATED.PNG"
    repo.add(hello_term)
    repo.commit()

    # Parent not changed.
    assert_parent_has(3, None, None, [])


def test_save_remove_parent_breaks_link(app_context, repo, hello_term):
    "Parent is left."
    hello_term.parents.append("parent")
    repo.add(hello_term)
    repo.commit()
    sql = "select WoTextLC from words"
    assert_sql_result(sql, ["hello", "parent"], "records exists")
    assert_record_count_equals("wordparents", 1, "linked")

    hello_term.parents.remove("parent")
    repo.add(hello_term)
    repo.commit()
    assert_sql_result(sql, ["hello", "parent"], "both stay")
    assert_record_count_equals("wordparents", 0, "link broken")


def test_save_change_parent_breaks_old_link(app_context, repo, hello_term):
    "Parent is left."
    hello_term.parents.append("parent")
    repo.add(hello_term)
    repo.commit()
    sql = "select WoTextLC from words order by WoTextLC"
    assert_sql_result(sql, ["hello", "parent"], "records exists")

    sqlparent = """
    select c.WoTextLC, p.WoTextLC
    from wordparents
    inner join words as c on WpWoID = c.WoID
    inner join words as p on WpParentWoID = p.WoID
    """
    assert_sql_result(sqlparent, ["hello; parent"], "records exists")

    hello_term.parents.remove("parent")
    hello_term.parents.append("new")
    repo.add(hello_term)
    repo.commit()
    assert_sql_result(sql, ["hello", "new", "parent"], "all stay")
    assert_sql_result(sqlparent, ["hello; new"], "changed")


def test_save_term_with_multiple_parents_unsets_sync_status(
    app_context, repo, hello_term
):
    "Ensure DB data is good on save."
    hello_term.parents.append("parent")
    hello_term.sync_status = True
    repo.add(hello_term)
    repo.commit()

    sql = "select WoTextLC, WoSyncStatus from words where WoTextLC = 'hello'"
    assert_sql_result(sql, ["hello; 1"], "link set")

    hello_term.parents.append("parent2")
    hello_term.sync_status = True
    repo.add(hello_term)
    repo.commit()
    assert_sql_result(sql, ["hello; 0"], "link UN-set")


def test_cant_set_term_as_its_own_parent(app_context, repo, hello_term):
    """
    Would create obvious circular ref
    """
    hello_term.parents = [hello_term.text]
    repo.add(hello_term)
    repo.commit()
    assert_sql_result("select WoText from words", ["HELLO"], "term entered")
    assert_sql_result("select * from wordparents", [], "no records")


@pytest.fixture(name="_existing_parent")
def fixture_existing_parent(app_context, english):
    "Create an existing parent."
    term = DBTerm(english, "parent")
    db.session.add(term)
    db.session.commit()
    return term


def test_save_with_existing_parent_creates_link(
    app_context, repo, hello_term, _existing_parent
):
    """
    new parent link is created, and is assigned translation and image and tag.
    """

    sql = "select WoID, WoText from words"
    assert_sql_result(sql, ["1; parent"], "initial state")

    hello_term.parents = ["parent"]
    hello_term.term_tags = ["a", "b"]
    dbterm = repo.add(hello_term)
    repo.commit()

    assert_sql_result(sql, ["1; parent", "2; HELLO"], "new record added")
    assert len(dbterm.parents) == 1
    assert dbterm.parents[0].text == "parent"


def test_save_existing_parent_gets_translation_if_missing(
    app_context, repo, hello_term, _existing_parent
):
    """
    Translation and image applied if missing.
    """

    sql = "select WoID, WoText, WoTranslation from words"
    assert_sql_result(sql, ["1; parent; None"], "initial state")

    hello_term.parents = ["parent"]
    hello_term.term_tags = ["a", "b"]
    repo.add(hello_term)
    repo.commit()

    assert_sql_result(
        sql, ["1; parent; greeting", "2; HELLO; greeting"], "transl saved"
    )


## Find tests.


def test_load(empty_db, english, repo):
    "Smoke test."
    t = DBTerm(english, "Hello")
    t.set_current_image("hello.png")
    t.add_term_tag(TermTag("a"))
    t.add_term_tag(TermTag("b"))
    db.session.add(t)
    db.session.commit()
    assert t.id is not None, "ID assigned"

    term = repo.load(t.id)
    assert term.id == t.id
    assert term.language_id == t.language.id
    assert term.language_id == english.id
    assert term.text == "Hello"
    assert term.original_text == "Hello"
    assert term.current_image == "hello.png"
    assert len(term.parents) == 0
    assert term.term_tags == ["a", "b"]


def test_load_throws_if_bad_id(app_context, repo):
    "If app says term id = 7 exists, and it doesn't, that's a problem."
    with pytest.raises(ValueError):
        repo.load(9876)


def test_find_is_found(spanish, app_context, repo):
    "Find by text finds regardless of case."
    add_terms(spanish, ["PARENT"])
    cases = ["PARENT", "parent", "pAReNt"]
    for c in cases:
        p = repo.find(spanish.id, c)
        assert p is not None, f"parent found for case {c}"
        assert p.text == "PARENT", f"parent found for case {c}"


def test_find_not_found_returns_none(spanish, repo):
    "No match = none."
    p = repo.find(spanish.id, "unknown_term")
    assert p is None, "nothing found"


def test_find_only_looks_in_specified_language(spanish, english, repo):
    "Only search language."
    add_terms(english, ["hola"])
    p = repo.find(spanish.id, "hola")
    assert p is None, "english terms not checked"


def test_find_existing_multi_word(spanish, repo):
    """
    Domain objects DO have zero-width strings in them.

    This is necessary to disambiguate terms such as
    "集めれ" and "集め/れ", both of which are valid
    parser results (though perhaps not *useful* parser
    results, but that's a different question).
    """
    add_terms(spanish, ["una bebida"])
    zws = "\u200B"
    term_with_zws = f"una{zws} {zws}bebida"
    t = repo.find(spanish.id, term_with_zws)
    assert t.id > 0
    assert t.text == term_with_zws

    t = repo.find(spanish.id, "una bebida")
    assert t.id > 0
    assert t.text == term_with_zws


## Find or new tests.


def test_find_or_new_existing_word(spanish, repo):
    "Existing term is found."
    add_terms(spanish, ["BEBIDA"])
    t = repo.find_or_new(spanish.id, "bebida")
    assert t.id > 0, "exists"
    assert t.text == "BEBIDA"
    assert t.language_id == spanish.id, "lang id set"


def test_find_or_new_non_existing(spanish, repo):
    "Returns new term."
    t = repo.find_or_new(spanish.id, "TENGO")
    assert t.id is None
    assert t.text == "TENGO"
    assert t.language_id == spanish.id, "lang id set"


def test_find_or_new_existing_multi_word(spanish, repo):
    "Spaces etc handled correctly."
    add_terms(spanish, ["una bebida"])
    zws = "\u200B"
    term_with_zws = f"una{zws} {zws}bebida"
    t = repo.find_or_new(spanish.id, f"una{zws} {zws}bebida")
    assert t.id > 0
    assert t.text == term_with_zws

    t = repo.find_or_new(spanish.id, "una bebida")
    assert t.id > 0
    assert t.text == term_with_zws


def test_find_or_new_new_multi_word(spanish, repo):
    "ZWS added correctly."
    zws = "\u200B"
    t = repo.find_or_new(spanish.id, f"una{zws} {zws}bebida")
    assert t.id is None
    assert t.text == f"una{zws} {zws}bebida"

    t = repo.find_or_new(spanish.id, "una bebida")
    assert t.id is None
    assert t.text == f"una{zws} {zws}bebida"


def test_find_or_new_ambiguous_japanese_terms(japanese, repo):
    """
    Characterization test only: behaviour of find_or_new for
    ambiguously parsable 集めれ terms

    See comments in find_or_new for notes.
    """
    s = "集めれ"
    term = DBTerm.create_term_no_parsing(japanese, s)
    db.session.add(term)
    db.session.commit()

    t = repo.find_or_new(japanese.id, s)
    assert t.id is None, "do _not_ have term, searching for string without context"
    zws = "\u200B"
    assert t.text == f"集め{zws}れ", "returns a new term"


## Matches tests.


@pytest.fixture(name="_multiple_terms")
def fixture_multiple(english, spanish, app_context):
    "Create multiple terms for find_matches tests."
    add_terms(english, ["parent"])
    add_terms(spanish, ["parent", "pare", "gato", "tengo uno"])


@pytest.mark.find_match
def test_find_matches_only_returns_language_matches(spanish, repo, _multiple_terms):
    "Searches match the start of string."
    for c in ["PARE", "pare", "PAR"]:
        matches = repo.find_matches(spanish.id, c)
        assert len(matches) == 2, c
        assert matches[0].text == "pare"
        assert matches[0].language_id == spanish.id, "language included"


@pytest.mark.find_match
def test_find_matches_returns_empty_if_no_match_or_empty_string(
    spanish, repo, _multiple_terms
):
    "Empty if no match."
    for c in ["", "x"]:
        matches = repo.find_matches(spanish.id, c)
        assert len(matches) == 0, c


@pytest.mark.find_match
def test_find_matches_multiword_respects_zws(spanish, repo, _multiple_terms):
    "zws handled correctly in search."
    zws = "\u200B"
    for c in [f"tengo{zws} {zws}uno", "tengo uno", "tengo"]:
        matches = repo.find_matches(spanish.id, c)
        assert len(matches) == 1, f'have match for case "{c}"'
        assert matches[0].text == f"tengo{zws} {zws}uno"


def assert_find_matches_returns(term_repo, spanish, s, expected):
    "Helper, assert returns expected content."
    matches = term_repo.find_matches(spanish.id, s)
    actual = ", ".join([t.text for t in matches])
    assert ", ".join(expected) == actual


def test_find_matches_starting_matches_sort_to_top(spanish, repo):
    "All matches return, but words that start with the same thing sort to top."
    add_terms(spanish, ["tener", "contener", "sostener", "ten", "contengo", "xxx"])
    assert_find_matches_returns(
        repo, spanish, "ten", ["ten", "tener", "contener", "contengo", "sostener"]
    )
    assert_find_matches_returns(repo, spanish, "ene", ["contener", "sostener", "tener"])
    assert_find_matches_returns(repo, spanish, "x", ["xxx"])


@pytest.mark.find_match
def test_find_matches_initial_check(spanish, repo):
    "Searches match any part of string."
    add_terms(spanish, ["abc", "abcd", "bcd"])

    assert_find_matches_returns(repo, spanish, "ab", ["abc", "abcd"])
    assert_find_matches_returns(repo, spanish, "abcd", ["abcd"])
    assert_find_matches_returns(repo, spanish, "bcd", ["bcd", "abcd"])
    assert_find_matches_returns(repo, spanish, "yy", [])


@pytest.mark.find_match
def test_find_matches_terms_with_children_go_to_top(spanish, repo):
    """
    Parents go to the top, then the rest ... but exact matches trumps parent.
    """
    terms = [("abc", "abcParent"), ("axy", "axyParent")]
    for term, parent in terms:
        t = Term()
        t.language = spanish
        t.language_id = spanish.id
        t.text = term
        t.parents.append(parent)
        repo.add(t)
    repo.commit()

    assert_find_matches_returns(
        repo, spanish, "a", ["abcParent", "axyParent", "abc", "axy"]
    )
    assert_find_matches_returns(repo, spanish, "ab", ["abcParent", "abc"])
    assert_find_matches_returns(repo, spanish, "abc", ["abc", "abcParent"])



"""
MandarinParser tests.
"""

import tempfile
import os
import pytest

# pylint: disable=wrong-import-order
from lute.models.term import Term
from lute.parse.base import ParsedToken

from lute_mandarin_parser.parser import MandarinParser


@pytest.mark.parametrize(
    "text,expected_count",
    [
        ("我", 1),
        ("运气", 1),
        ("你说", 2),
        ("我不相信", 3),
        ("我冒了严寒 ，回到相隔二千馀里，别了二十馀年的故乡去。", 21),
    ],
)
def test_token_count(text, expected_count, mandarin_chinese):
    """
    token_count checks.
    """
    t = Term(mandarin_chinese, text)
    assert t.token_count == expected_count, text
    assert t.text_lc == t.text


def assert_tokens_equals(text, lang, expected):
    """
    Parsing a text using a language should give the expected parsed tokens.

    expected is given as array of:
    [ original_text, is_word, is_end_of_sentence ]
    """
    p = MandarinParser()
    actual = p.get_parsed_tokens(text, lang)
    expected = [ParsedToken(*a) for a in expected]
    assert [str(a) for a in actual] == [str(e) for e in expected]


def test_end_of_sentence_stored_in_parsed_tokens(mandarin_chinese):
    """
    ParsedToken is marked as EOS=True at ends of sentences.
    """
    s = "你好。吃饭了吗？现在是2024年。"

    expected = [
        ("你好", True),
        ("。", False, True),
        ("吃饭", True),
        ("了", True),
        ("吗", True),
        ("？", False, True),
        ("现在", True),
        ("是", True),
        ("2024", False, False),
        ("年", True),
        ("。", False, True),
    ]
    assert_tokens_equals(s, mandarin_chinese, expected)


def test_carriage_returns_treated_as_reverse_p_character(mandarin_chinese):
    """
    Returns need to be marked with the backwards P for rendering etc.
    """
    s = "你好。\n现在。"

    expected = [
        ("你好", True),
        ("。", False, True),
        ("¶", False, True),
        ("现在", True),
        ("。", False, True),
    ]
    assert_tokens_equals(s, mandarin_chinese, expected)


def test_readings():
    """
    Parser returns readings if they add value.
    """
    p = MandarinParser()

    no_reading = ["Hello"]  # roman  # only katakana  # only hiragana

    for c in no_reading:
        assert p.get_reading(c) is None, c

    cases = [("你好", "nǐ hǎo"), ("欢迎", "huān yíng"), ("中国", "zhōng guó")]

    for c in cases:
        assert p.get_reading(c[0]) == c[1], c[0]


@pytest.fixture(name="_datadir")
def datadir():
    "The data_directory for the plugin."
    with tempfile.TemporaryDirectory() as temp_dir:
        MandarinParser.data_directory = temp_dir
        MandarinParser.init_data_directory()
        exceptions_file = os.path.join(temp_dir, "parser_exceptions.txt")
        assert os.path.exists(exceptions_file), "File should exist after init"
        yield


def test_term_found_in_exceptions_file_is_split(mandarin_chinese, _datadir):
    "User can specify parsing exceptions in file."
    s = "清华大学"

    def parsed_tokens():
        p = MandarinParser()
        return [t.token for t in p.get_parsed_tokens(s, mandarin_chinese)]

    assert ["清华大学"] == parsed_tokens(), "No exceptions"

    exceptions_file = MandarinParser.parser_exceptions_file()
    assert os.path.exists(exceptions_file), "Sanity check only."

    def set_parse_exceptions(array_of_exceptions):
        with open(exceptions_file, "w", encoding="utf8") as ef:
            ef.write("\n".join(array_of_exceptions))

    set_parse_exceptions(["清华大学"])
    assert ["清华大学"] == parsed_tokens(), "mapped to self"

    set_parse_exceptions(["清华,大学"])
    assert ["清华", "大学"] == parsed_tokens(), "Exceptions consulted during parse"

    set_parse_exceptions(["清华,大,学"])
    assert ["清华", "大", "学"] == parsed_tokens(), "multiple splits tokens"

    set_parse_exceptions(["清华,大,学", "清华,大学"])
    assert ["清华", "大学"] == parsed_tokens(), "Last rule takes precedence"

    set_parse_exceptions(["清学"])
    assert ["清华大学"] == parsed_tokens(), "no match = parsed as-is"

    set_parse_exceptions(["清华,大学", "大,学"])
    assert ["清华", "大", "学"] == parsed_tokens(), "Recursive splitting"

    set_parse_exceptions(["大,学", "清华,大学"])
    assert ["清华", "大", "学"] == parsed_tokens(), "Order doesn't matter"

    set_parse_exceptions(["清华, 大学", " 大 ,  学 "])
    assert ["清华", "大", "学"] == parsed_tokens(), "Spaces are ignored"



"""
Smoke test get post data.
"""

import json
from lute.models.term import Term, TermTag
from lute.db import db
from lute.models.srsexport import SrsExportSpec
from lute.ankiexport.service import Service


def test_smoke_get_post_data(empty_db, spanish):
    "Misc data check - parent and tags are saved."
    term = Term(spanish, "un gatito")
    term.translation = "t_trans\nt_extra"
    term.romanization = "t_rom"
    term.set_flash_message("hello")
    term.add_term_tag(TermTag("t_tag"))
    term.set_current_image("blah.jpg")

    parent = Term(spanish, "un gato")
    parent.translation = "p_trans\np_extra"
    parent.romanization = "p_rom"
    parent.add_term_tag(TermTag("p_tag"))

    component = Term(spanish, "gatito")
    component.translation = "c_trans\nc_extra"
    component.romanization = "c_rom"
    component.add_term_tag(TermTag("c_tag"))

    term.add_parent(parent)
    db.session.add(term)
    db.session.add(component)
    db.session.commit()

    term.set_flash_message("hello")

    db.session.add(term)
    db.session.add(parent)
    db.session.add(component)
    db.session.commit()

    spec = SrsExportSpec()
    spec.id = 1
    spec.export_name = "export_name"
    spec.criteria = ""
    spec.deck_name = "good_deck"
    spec.note_type = "good_note"
    spec.field_mapping = json.dumps({"a": "{ language }"})
    spec.active = True
    spec.field_mapping = json.dumps(
        {
            "a": "{ language }",
            "b": "{ image }",
            "c": "{ term }",
            "d": "{ sentence }",
        }
    )

    anki_decks = ["good_deck"]
    anki_notes = {"good_note": ["a", "b", "c", "d"]}
    svc = Service(anki_decks, anki_notes, [spec])
    post_data = svc.get_ankiconnect_post_data([term.id], {}, "dummyurl", db.session)
    # print(post_data, flush=True)

    expected = {
        1: {
            "export_name": {
                "action": "multi",
                "params": {
                    "actions": [
                        {
                            "action": "storeMediaFile",
                            "params": {
                                "filename": "LUTE_TERM_1.jpg",
                                "url": "dummyurl/userimages/1/blah.jpg",
                            },
                        },
                        {
                            "action": "addNote",
                            "params": {
                                "note": {
                                    "deckName": "good_deck",
                                    "modelName": "good_note",
                                    "fields": {
                                        "a": "Spanish",
                                        "b": '<img src="LUTE_TERM_1.jpg">',
                                        "c": "un gatito",
                                    },
                                    "tags": ["lute", "p_tag", "t_tag"],
                                }
                            },
                        },
                    ]
                },
            }
        }
    }

    assert post_data == expected, "Got post data"



"""Field to value mapper.

e.g. given dict like

{
  "lute_term_id": "{ id }",
  "term": "{ term }",
  "tags": "{ tags:["masc", "fem"] }"
}

extracts data from the given term and generates a mapping of field to
actual values to send to AnkiConnect.
"""

import re
import pyparsing as pp
from pyparsing import (
    quotedString,
    Suppress,
)
from pyparsing.exceptions import ParseException
from lute.models.term import Term
from lute.models.language import Language
from lute.ankiexport.exceptions import AnkiExportConfigurationError


class SentenceLookup:
    "Sentence lookup, finds in a supplied dictionary or from db."

    def __init__(self, default_sentences_by_term_id, references_repo):
        "init"
        sdict = {}
        for k, v in default_sentences_by_term_id.items():
            sdict[int(k)] = v
        self.default_sentences_by_term_id = sdict
        self.references_repo = references_repo

    def get_sentence_for_term(self, term_id):
        "Get sentence from the dict, or do a lookup."
        tid = int(term_id)
        if tid in self.default_sentences_by_term_id:
            return self.default_sentences_by_term_id[tid]

        refs = self.references_repo.find_references_by_id(term_id)
        term_refs = refs["term"] or []
        if len(term_refs) == 0:
            return ""
        return term_refs[0].sentence


def _all_terms(term):
    "Term and any parents."
    ret = [term]
    ret.extend(term.parents)
    return ret


def _all_tags(term):
    "Tags for term and all parents."
    ret = [tt.text for t in _all_terms(term) for tt in t.term_tags]
    return sorted(list(set(ret)))


def get_values_and_media_mapping(term, sentence_lookup, mapping):
    """
    Get the value replacements to be put in the mapping, and build
    dict of new filenames to original filenames.
    """

    def all_translations():
        ret = [term.translation or ""]
        for p in term.parents:
            if p.translation not in ret:
                ret.append(p.translation or "")
        return [r for r in ret if r.strip() != ""]

    def parse_keys_needing_calculation(calculate_keys, media_mappings):
        """
        Build a parser for some keys in the mapping string, return
        calculated value to use in the mapping.  SIDE EFFECT:
        adds ankiconnect post actions to post_actions if needed
        (e.g. for image uploads).

        e.g. the mapping "article: { tags:["der", "die", "das"] }"
        needs to be parsed to extract certain tags from the current
        term.
        """

        def get_filtered_tags(tagvals):
            "Get term tags matching the list."
            # tagvals is a pyparsing ParseResults, use list() to convert to strings.
            ftags = [tt for tt in _all_tags(term) if tt in list(tagvals)]
            return ", ".join(ftags)

        def handle_image(_):
            id_images = [
                (t, t.get_current_image())
                for t in _all_terms(term)
                if t.get_current_image() is not None
            ]
            image_srcs = []
            for t, imgfilename in id_images:
                new_filename = f"LUTE_TERM_{t.id}.jpg"
                image_url = f"/userimages/{t.language.id}/{imgfilename}"
                media_mappings[new_filename] = image_url
                image_srcs.append(f'<img src="{new_filename}">')

            return "".join(image_srcs)

        def handle_sentences(_):
            "Get sample sentence for term."
            if term.id is None:
                # Dummy parse.
                return ""
            return sentence_lookup.get_sentence_for_term(term.id)

        quotedString.setParseAction(pp.removeQuotes)
        tag_matcher = (
            Suppress("tags")
            + Suppress(":")
            + Suppress("[")
            + pp.delimitedList(quotedString)
            + Suppress("]")
        )
        image = Suppress("image")
        sentence = Suppress("sentence")

        matcher = (
            tag_matcher.set_parse_action(get_filtered_tags)
            | image.set_parse_action(handle_image)
            | sentence.set_parse_action(handle_sentences)
        )

        calc_replacements = {
            # Matchers return the value that should be used as the
            # replacement value for the given mapping string.  e.g.
            # tags["der", "die"] returns "der" if term.tags = ["der", "x"]
            k: matcher.parseString(k).asList()[0]
            for k in calculate_keys
        }

        return calc_replacements

    def remove_zws(replacements_dict):
        cleaned = {}
        for key, value in replacements_dict.items():
            if isinstance(value, str):
                cleaned[key] = value.replace("\u200B", "")
            else:
                cleaned[key] = value
        return cleaned

    # One-for-one replacements in the mapping string.
    # e.g. "{ id }" is replaced by term.termid.
    replacements = {
        "id": term.id,
        "term": term.text,
        "language": term.language.name,
        "parents": ", ".join([p.text for p in term.parents]),
        "tags": ", ".join(_all_tags(term)),
        "translation": "<br>".join(all_translations()),
    }

    mapping_string = "; ".join(mapping.values())
    calc_keys = [
        k
        for k in set(re.findall(r"{\s*(.*?)\s*}", mapping_string))
        if k not in replacements
    ]

    media_mappings = {}
    calc_replacements = parse_keys_needing_calculation(calc_keys, media_mappings)

    final_replacements = {**replacements, **calc_replacements}
    cleaned = remove_zws(final_replacements)
    return (cleaned, media_mappings)


def validate_mapping(mapping):
    "Check mapping with a dummy Term."
    t = Term(Language(), "")
    refsrepo = None
    try:
        get_values_and_media_mapping(t, refsrepo, mapping)
    except ParseException as ex:
        msg = f'Invalid field mapping "{ex.line}"'
        raise AnkiExportConfigurationError(msg) from ex


def get_fields_and_final_values(mapping, replacements):
    "Break mapping string into fields, apply replacements."
    ret = {}
    for fieldname, value in mapping.items():
        subbed_value = value
        for k, v in replacements.items():
            pattern = rf"{{\s*{re.escape(k)}\s*}}"
            subbed_value = re.sub(pattern, f"{v}", subbed_value)
        if subbed_value.strip() != "":
            ret[fieldname.strip()] = subbed_value.strip()
    return ret



"""
SrsExportSpec form.
"""

from flask_wtf import FlaskForm
from wtforms import StringField, SelectField, BooleanField, TextAreaField, HiddenField
from wtforms.validators import DataRequired, Length
from lute.ankiexport.service import Service
from lute.models.srsexport import SrsExportSpec


class SrsExportSpecForm(FlaskForm):
    "Srs export spec."

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Flask doesn't allow for "general" errors, so handle those
        # specially.
        self.general_errors = []

        # Extra data added to the form during POST
        # to simplify validation.
        self.anki_deck_names = None
        self.anki_note_types = None

    export_name = StringField(
        "Export Name", validators=[DataRequired(), Length(max=200)]
    )
    criteria = TextAreaField(
        "Criteria", render_kw={"placeholder": "Leave blank to always export"}
    )
    deck_name = SelectField("Deck Name", validators=[DataRequired(), Length(max=200)])
    note_type = SelectField("Note Type", validators=[DataRequired(), Length(max=200)])
    field_mapping = HiddenField(
        "Field Mapping", validators=[DataRequired(), Length(max=1000)]
    )
    active = BooleanField("Active", default=True)

    def validate(self, extra_validators=None):
        """Custom validation logic."""
        if not super().validate(extra_validators):
            return False  # Return early if standard validation fails

        temp_spec = SrsExportSpec()
        self.populate_obj(temp_spec)
        svc = Service(self.anki_deck_names, self.anki_note_types, [temp_spec])
        self.general_errors = svc.validate_spec(temp_spec)
        if len(self.general_errors) > 0:
            return False

        return True



"""
Book service tests.
"""

import os
from contextlib import ExitStack
from lute.db import db
from lute.models.repositories import BookRepository
from lute.book.model import Book
from lute.book.service import Service


def get_test_files():
    "Return test files pair."
    thisdir = os.path.dirname(os.path.realpath(__file__))
    sample_files = os.path.join(thisdir, "..", "..", "acceptance", "sample_files")
    text_path = os.path.join(sample_files, "hola.txt")
    mp3_path = os.path.join(sample_files, "fake.mp3")
    with open(text_path, "r", encoding="utf-8") as fp:
        assert fp.read().strip() == "Tengo un amigo.", "Sanity check only."
    with open(mp3_path, "r", encoding="utf-8") as fp:
        assert fp.read().strip() == "fake mp3 file", "Sanity check only."
    return (text_path, mp3_path)


def test_create_book_from_file_paths(app, app_context, spanish):
    "Create a book using the DTO, to be populated by the form."
    text_path, mp3_path = get_test_files()

    b = Book()
    b.title = "Hola"
    b.language_id = spanish.id
    b.text_source_path = text_path
    b.audio_source_path = mp3_path

    svc = Service()
    svc.import_book(b, db.session)

    repo = BookRepository(db.session)
    book = repo.find_by_title("Hola", spanish.id)
    assert book.title == "Hola", "title"
    assert book.texts[0].text == "Tengo un amigo.", "Got content"

    assert book.audio_filename is not None, "Have audio file"
    assert book.audio_filename.endswith("mp3"), "still an mp3"
    useraudiopath = app.env_config.useraudiopath
    full_audio_path = os.path.join(useraudiopath, book.audio_filename)
    assert os.path.exists(full_audio_path), "file saved"

    with open(full_audio_path, "r", encoding="utf-8") as fp:
        assert fp.read().strip() == "fake mp3 file", "correct content copied."


def test_create_book_from_streams(app, app_context, spanish):
    "Create a book using streams, as given by the form."
    text_path, mp3_path = get_test_files()

    b = Book()
    b.title = "Hola"
    b.language_id = spanish.id
    with ExitStack() as stack:
        b.text_stream = stack.enter_context(open(text_path, mode="rb"))
        b.text_stream_filename = "blah.txt"
        b.audio_stream = stack.enter_context(open(mp3_path, mode="rb"))
        b.audio_stream_filename = "blah.mp3"
        svc = Service()
        svc.import_book(b, db.session)

    repo = BookRepository(db.session)
    book = repo.find_by_title("Hola", spanish.id)
    assert book.title == "Hola", "title"
    assert book.texts[0].text == "Tengo un amigo.", "Got content"

    assert book.audio_filename is not None, "Have audio file"
    assert book.audio_filename.endswith("mp3"), "still an mp3"
    useraudiopath = app.env_config.useraudiopath
    full_audio_path = os.path.join(useraudiopath, book.audio_filename)
    assert os.path.exists(full_audio_path), "file saved"
    with open(full_audio_path, "r", encoding="utf-8") as fp:
        assert fp.read().strip() == "fake mp3 file", "correct content copied."



"""
Service, validates and posts.
"""

import json
from lute.models.repositories import TermRepository
from lute.term.model import ReferencesRepository
from lute.ankiexport.exceptions import AnkiExportConfigurationError
from lute.ankiexport.field_mapping import (
    get_values_and_media_mapping,
    validate_mapping,
    get_fields_and_final_values,
    SentenceLookup,
)
from lute.ankiexport.criteria import (
    evaluate_criteria,
    validate_criteria,
)


class Service:
    "Srs export service."

    def __init__(
        self,
        anki_deck_names,
        anki_note_types_and_fields,
        export_specs,
    ):
        "init"
        self.anki_deck_names = anki_deck_names
        self.anki_note_types_and_fields = anki_note_types_and_fields
        self.export_specs = export_specs

    def validate_spec(self, spec):
        """
        Returns array of errors if any for the given spec.
        """
        if not spec.active:
            return []

        errors = []

        try:
            validate_criteria(spec.criteria)
        except AnkiExportConfigurationError as ex:
            errors.append(str(ex))

        if spec.deck_name not in self.anki_deck_names:
            errors.append(f'No deck name "{spec.deck_name}"')

        valid_note_type = spec.note_type in self.anki_note_types_and_fields
        if not valid_note_type:
            errors.append(f'No note type "{spec.note_type}"')

        mapping = None
        try:
            mapping = json.loads(spec.field_mapping)
        except json.decoder.JSONDecodeError:
            errors.append("Mapping is not valid json")

        if valid_note_type and mapping:
            note_fields = self.anki_note_types_and_fields.get(spec.note_type, {})
            bad_fields = [f for f in mapping.keys() if f not in note_fields]
            if len(bad_fields) > 0:
                bad_fields = ", ".join(bad_fields)
                msg = f"Note type {spec.note_type} does not have field(s): {bad_fields}"
                errors.append(msg)

        if mapping:
            try:
                validate_mapping(json.loads(spec.field_mapping))
            except AnkiExportConfigurationError as ex:
                errors.append(str(ex))

        return errors

    def validate_specs(self):
        """
        Return hash of spec ids and any config errors.
        """
        failures = {}
        for spec in self.export_specs:
            v = self.validate_spec(spec)
            if len(v) != 0:
                failures[spec.id] = "; ".join(v)
        return failures

    def validate_specs_failure_message(self):
        "Failure message for alerts."
        failures = self.validate_specs()
        msgs = []
        for k, v in failures.items():
            spec = next(s for s in self.export_specs if s.id == k)
            msgs.append(f"{spec.export_name}: {v}")
        return msgs

    def _all_terms(self, term):
        "Term and any parents."
        ret = [term]
        ret.extend(term.parents)
        return ret

    def _all_tags(self, term):
        "Tags for term and all parents."
        ret = [tt.text for t in self._all_terms(term) for tt in t.term_tags]
        return sorted(list(set(ret)))

    # pylint: disable=too-many-arguments,too-many-positional-arguments
    def _build_ankiconnect_post_json(
        self,
        mapping,
        media_mappings,
        lute_and_term_tags,
        deck_name,
        model_name,
    ):
        "Build post json for term using the mappings."

        post_actions = []
        for new_filename, original_url in media_mappings.items():
            hsh = {
                "action": "storeMediaFile",
                "params": {
                    "filename": new_filename,
                    "url": original_url,
                },
            }
            post_actions.append(hsh)

        post_actions.append(
            {
                "action": "addNote",
                "params": {
                    "note": {
                        "deckName": deck_name,
                        "modelName": model_name,
                        "fields": mapping,
                        "tags": lute_and_term_tags,
                    }
                },
            }
        )

        return {"action": "multi", "params": {"actions": post_actions}}

    def get_ankiconnect_post_data_for_term(self, term, base_url, sentence_lookup):
        """
        Get post data for a single term.
        This assumes that all the specs are valid!
        Separate method for unit testing.
        """
        use_exports = [
            spec
            for spec in self.export_specs
            if spec.active and evaluate_criteria(spec.criteria, term)
        ]
        # print(f"Using {len(use_exports)} exports")

        ret = {}
        for export in use_exports:
            mapping = json.loads(export.field_mapping)
            replacements, mmap = get_values_and_media_mapping(
                term, sentence_lookup, mapping
            )
            for k, v in mmap.items():
                mmap[k] = base_url + v
            updated_mapping = get_fields_and_final_values(mapping, replacements)
            tags = ["lute"] + self._all_tags(term)

            p = self._build_ankiconnect_post_json(
                updated_mapping,
                mmap,
                tags,
                export.deck_name,
                export.note_type,
            )
            ret[export.export_name] = p

        return ret

    def get_ankiconnect_post_data(
        self, term_ids, termid_sentences, base_url, db_session
    ):
        """
        Build data to be posted.

        Throws if any validation failure or mapping failure, as it's
        annoying to handle partial failures.
        """

        msgs = self.validate_specs_failure_message()
        if len(msgs) > 0:
            show_msgs = [f"* {m}" for m in msgs]
            show_msgs = "\n".join(show_msgs)
            err_msg = "Anki export configuration errors:\n" + show_msgs
            raise AnkiExportConfigurationError(err_msg)

        repo = TermRepository(db_session)

        refsrepo = ReferencesRepository(db_session)
        sentence_lookup = SentenceLookup(termid_sentences, refsrepo)

        ret = {}
        for tid in term_ids:
            term = repo.find(tid)
            pd = self.get_ankiconnect_post_data_for_term(
                term, base_url, sentence_lookup
            )
            if len(pd) > 0:
                ret[tid] = pd

        return ret



"""
Term service apply_ajax_update tests.

Generally smoke tests.
"""

import pytest
from lute.models.repositories import TermRepository
from lute.db import db
from lute.term.service import Service, TermServiceException
from tests.utils import add_terms
from tests.dbasserts import assert_sql_result

# pylint: disable=unbalanced-tuple-unpacking,missing-function-docstring


def assert_updated(termid, expected, msg=""):
    "Return a term as a string for fast asserts."
    t = TermRepository(db.session).find(termid)
    bad_keys = [
        k
        for k in expected.keys()
        if k not in ["parents", "status", "tags", "translation"]
    ]
    assert len(bad_keys) == 0, "no bad keys " + ", ".join(bad_keys)
    if "translation" in expected:
        assert expected["translation"] == t.translation, msg
    if "parents" in expected:
        parents = expected["parents"]
        assert parents == sorted([p.text for p in t.parents]), msg
        should_by_sync = len(parents) == 1
        assert t.sync_status == should_by_sync, f"sync status with parents = {parents}"
    if "status" in expected:
        assert expected["status"] == t.status, msg
    if "tags" in expected:
        assert expected["tags"] == sorted([tag.text for tag in t.term_tags]), msg


def _apply_updates(term_id, update_type, values):
    "Apply updates."
    svc = Service(db.session)
    svc.apply_ajax_update(term_id, update_type, values)


def test_smoke_test(app_context, spanish):
    "Smoke test."
    [t, _] = add_terms(spanish, ["T", "P"])
    _apply_updates(t.id, "parents", ["P", "X"])
    _apply_updates(t.id, "status", 4)
    _apply_updates(t.id, "term_tags", ["tag1", "tag2"])
    _apply_updates(t.id, "translation", "new_translation")
    expected = {
        "translation": "new_translation",
        "parents": ["P", "X"],
        "status": 4,
        "tags": ["tag1", "tag2"],
    }
    assert_updated(t.id, expected, "smoke all items")


def test_term_gets_parent_status_if_parent_status_set(app_context, spanish):
    "Smoke test."
    [t, p] = add_terms(spanish, ["T", "P"])
    t.status = 1
    p.status = 4
    db.session.add(t)
    db.session.add(p)
    db.session.commit()

    _apply_updates(t.id, "parents", ["P"])
    expected = {"status": 4}
    assert_updated(t.id, expected, "gets parent's status")
    assert_updated(p.id, expected, "parent keeps status")


def test_term_keeps_status_if_parent_status_unknown(app_context, spanish):
    "Smoke test."
    [t, p] = add_terms(spanish, ["T", "P"])
    t.status = 1
    p.status = 0
    db.session.add(t)
    db.session.add(p)
    db.session.commit()

    _apply_updates(t.id, "parents", ["P"])
    expected = {"status": 1}
    assert_updated(t.id, expected, "keeps own status")
    expected = {"status": 1}
    assert_updated(p.id, expected, "parent gets child's status!")


def test_bad_term_id_throws(app_context):
    svc = Service(db.session)
    with pytest.raises(TermServiceException, match="No term with id -99"):
        svc.apply_ajax_update(-99, "status", 99)


def test_bad_update_type_throws(app_context, spanish):
    svc = Service(db.session)
    [t] = add_terms(spanish, ["T"])
    with pytest.raises(TermServiceException, match="Bad update type"):
        svc.apply_ajax_update(t.id, "trash", 99)


def test_bad_status_throws(app_context, spanish):
    svc = Service(db.session)
    [t] = add_terms(spanish, ["T"])
    with pytest.raises(TermServiceException, match="Bad status value"):
        svc.apply_ajax_update(t.id, "status", 42)


def test_can_remove_values(app_context, spanish):
    "Smoke test."
    [t, _] = add_terms(spanish, ["T", "P"])
    _apply_updates(t.id, "parents", ["P", "X"])
    _apply_updates(t.id, "term_tags", ["tag1", "tag2"])
    _apply_updates(t.id, "translation", "new_translation")
    expected = {
        "translation": "new_translation",
        "parents": ["P", "X"],
        "tags": ["tag1", "tag2"],
    }
    assert_updated(t.id, expected, "smoke all items")

    _apply_updates(t.id, "parents", [])
    _apply_updates(t.id, "term_tags", [])
    _apply_updates(t.id, "translation", "")
    expected = {
        "translation": None,
        "parents": [],
        "tags": [],
    }
    assert_updated(t.id, expected, "smoke all items updated back to nothing")


def test_parent_created_if_needed(app_context, spanish):
    [t, _] = add_terms(spanish, ["T", "P"])
    sql = "select WoText from words order by WoText"
    assert_sql_result(sql, ["P", "T"], "initial state")
    _apply_updates(t.id, "parents", ["P", "X"])
    assert_sql_result(sql, ["P", "T", "X"], "X created")


def test_correct_parent_found_if_has_zero_width_spaces(app_context, spanish):
    [t, _] = add_terms(spanish, ["T", "P Q"])
    sql = "select WoText from words order by WoText"
    assert_sql_result(sql, ["P/ /Q", "T"], "initial state")

    zws = "\u200B"
    parent = f"P{zws} {zws}Q"
    _apply_updates(t.id, "parents", [parent])
    expected = {"parents": [parent]}
    assert_updated(t.id, expected, "correct zws-included parent found")
    assert_sql_result(sql, ["P/ /Q", "T"], "nothing new created")



"""
Book domain objects.
"""

from lute.models.book import BookTag, Book as DBBook, Text as DBText
from lute.models.repositories import (
    BookRepository,
    BookTagRepository,
    LanguageRepository,
)


def token_group_generator(tokens, group_type, threshold=500):
    """
    A generator that yields groups of ParsedTokens grouped by sentence or paragraph
    with each group containing at least the threshold number of tokens.
    """
    current_group = []
    buff = []

    def trim_paras(tok_array):
        "Remove para tokens from beginning and end."
        while tok_array and tok_array[0].is_end_of_paragraph:
            tok_array.pop(0)
        while tok_array and tok_array[-1].is_end_of_paragraph:
            tok_array.pop()
        return tok_array

    def _matches_group_delimiter(tok):
        if group_type == "sentences":
            return tok.is_end_of_sentence
        if group_type == "paragraphs":
            return tok.is_end_of_paragraph
        raise RuntimeError("Unhandled type " + group_type)

    for token in tokens:
        buff.append(token)
        if _matches_group_delimiter(token):
            current_group.extend(buff)
            # pylint: disable=consider-using-generator
            current_count = sum([1 for t in current_group if t.is_word])
            buff = []

            # Yield if threshold exceeded.
            # Remove the final paragreph marker if it's there, it's not needed.
            if current_count > threshold:
                current_group = trim_paras(current_group)
                yield current_group
                current_group = []

    # Add any remaining tokens
    if buff:
        current_group.extend(buff)
    current_group = trim_paras(current_group)
    if current_group:
        yield current_group


class Book:  # pylint: disable=too-many-instance-attributes
    """
    A book domain object, to create/edit lute.models.book.Books.

    Book language can be specified either by language_id, or
    language_name.  language_name is useful for loading books via
    scripts/api.  language_id takes precedence.
    """

    def __init__(self):
        self.id = None
        self.language_id = None
        self.language_name = None
        self.title = None
        self.text = None
        self.source_uri = None
        self.audio_filename = None
        self.audio_current_pos = None
        self.audio_bookmarks = None
        self.book_tags = []

        self.threshold_page_tokens = 250
        self.split_by = "paragraphs"

        # The source file used for the book text.
        # Overrides the self.text if not None.
        self.text_source_path = None

        self.text_stream = None
        self.text_stream_filename = None

        # The source file used for audio.
        self.audio_source_path = None

        self.audio_stream = None
        self.audio_stream_filename = None

    def __repr__(self):
        return f"<Book (id={self.id}, title='{self.title}')>"

    def add_tag(self, tag):
        self.book_tags.append(tag)


class Repository:
    """
    Maps Book BO to and from lute.model.Book.
    """

    def __init__(self, _session):
        self.session = _session
        self.book_repo = BookRepository(self.session)

    def load(self, book_id):
        "Loads a Book business object for the DBBook."
        dbb = self.book_repo.find(book_id)
        if dbb is None:
            raise ValueError(f"No book with id {book_id} found")
        return self._build_business_book(dbb)

    def find_by_title(self, book_title, language_id):
        "Loads a Book business object for the book with a given title."
        dbb = self.book_repo.find_by_title(book_title, language_id)
        if dbb is None:
            return None
        return self._build_business_book(dbb)

    def get_book_tags(self):
        "Get all available book tags, helper method."
        bts = self.session.query(BookTag).all()
        return sorted([t.text for t in bts])

    def add(self, book):
        """
        Add a book to be saved to the db session.
        Returns DBBook for tests and verification only,
        clients should not change it.
        """
        dbbook = self._build_db_book(book)
        self.session.add(dbbook)
        return dbbook

    def delete(self, book):
        """
        Delete.
        """
        if book.id is None:
            raise ValueError(f"book {book.title} not saved")
        b = self.book_repo.find(book.id)
        self.session.delete(b)

    def commit(self):
        """
        Commit everything.
        """
        self.session.commit()

    def _split_text_at_page_breaks(self, txt):
        "Break fulltext manually at lines consisting of '---' only."
        # Tried doing this with a regex without success.
        segments = []
        current_segment = ""
        for line in txt.split("\n"):
            if line.strip() == "---":
                segments.append(current_segment.strip())
                current_segment = ""
            else:
                current_segment += line + "\n"
        if current_segment:
            segments.append(current_segment.strip())
        return segments

    def _split_pages(self, book, language):
        "Split fulltext into pages, respecting sentences."

        pages = []
        for segment in self._split_text_at_page_breaks(book.text):
            tokens = language.parser.get_parsed_tokens(segment, language)
            for toks in token_group_generator(
                tokens, book.split_by, book.threshold_page_tokens
            ):
                s = "".join([t.token for t in toks])
                s = s.replace("\r", "").replace("¶", "\n")
                pages.append(s.strip())
        pages = [p for p in pages if p.strip() != ""]

        return pages

    def _build_db_book(self, book):
        "Convert a book business object to a DBBook."

        lang_repo = LanguageRepository(self.session)
        lang = None
        if book.language_id:
            lang = lang_repo.find(book.language_id)
        elif book.language_name:
            lang = lang_repo.find_by_name(book.language_name)
        if lang is None:
            msg = f"No language matching id={book.language_id} or name={book.language_name}"
            raise RuntimeError(msg)

        b = None
        if book.id is None:
            pages = self._split_pages(book, lang)
            b = DBBook(book.title, lang)
            for index, page in enumerate(pages):
                _ = DBText(b, page, index + 1)
        else:
            b = self.book_repo.find(book.id)

        b.title = book.title
        b.source_uri = book.source_uri
        b.audio_filename = book.audio_filename
        b.audio_current_pos = book.audio_current_pos
        b.audio_bookmarks = book.audio_bookmarks

        btr = BookTagRepository(self.session)
        booktags = []
        for s in book.book_tags:
            booktags.append(btr.find_or_create_by_text(s))
        b.remove_all_book_tags()
        for tt in booktags:
            b.add_book_tag(tt)

        return b

    def _build_business_book(self, dbbook):
        "Convert db book to Book."
        b = Book()
        b.id = dbbook.id
        b.language_id = dbbook.language.id
        b.language_name = dbbook.language.name
        b.title = dbbook.title
        b.text = None  # Not returning this for now
        b.source_uri = dbbook.source_uri
        b.audio_filename = dbbook.audio_filename
        b.audio_current_pos = dbbook.audio_current_pos
        b.audio_bookmarks = dbbook.audio_bookmarks
        b.book_tags = [t.text for t in dbbook.book_tags]
        return b



"""
User images routes.
"""

import os
from flask import Blueprint, send_from_directory, current_app

bp = Blueprint("userimages", __name__, url_prefix="/userimages")


@bp.route("/<int:lgid>/<path:f>", methods=["GET"])
def get_image(lgid, f):
    "Serve the image from the data/userimages directory."
    datapath = current_app.config["DATAPATH"]
    directory = os.path.join(datapath, "userimages", str(lgid))
    if not os.path.exists(os.path.join(directory, f)):
        return ""
    return send_from_directory(directory, f)



"""
All customizable user hotkeys.
"""

import pprint
import yaml


# Hotkeys and descriptions.
_ALL_HOTKEY_DATA = """
Copy:
- hotkey: hotkey_CopySentence
  desc: Copy the sentence of the current word
- hotkey: hotkey_CopyPara
  desc: Copy the paragraph of the current word
- hotkey: hotkey_CopyPage
  desc: Copy the full page

Misc:
- hotkey: hotkey_PageTermList
  desc: Show the term listing for the current page
- hotkey: hotkey_Bookmark
  desc: Bookmark the current page
- hotkey: hotkey_EditPage
  desc: Edit the current page
- hotkey: hotkey_NextTheme
  desc: Change to the next theme
- hotkey: hotkey_ToggleHighlight
  desc: Toggle highlights
- hotkey: hotkey_ToggleFocus
  desc: Toggle focus mode
- hotkey: hotkey_SaveTerm
  desc: Save term in term form
- hotkey: hotkey_PostTermsToAnki
  desc: Post selected terms to Anki

Navigation:
- hotkey: hotkey_StartHover
  desc: Deselect all words
- hotkey: hotkey_PrevWord
  desc: Move to previous word
- hotkey: hotkey_NextWord
  desc: Move to next word
- hotkey: hotkey_PrevUnknownWord
  desc: Move to previous unknown word
- hotkey: hotkey_NextUnknownWord
  desc: Move to next unknown word
- hotkey: hotkey_PrevSentence
  desc: Move to previous sentence
- hotkey: hotkey_NextSentence
  desc: Move to next sentence

Paging:
- hotkey: hotkey_PreviousPage
  desc: Go to previous page, do not mark current page read
- hotkey: hotkey_NextPage
  desc: Go to next page, do not mark current page read
- hotkey: hotkey_MarkReadWellKnown
  desc: Set remaining unknown words to Well Known, mark page as read, go to next page
- hotkey: hotkey_MarkRead
  desc: Mark page as read, go to next page
- hotkey: hotkey_MarkReadWellKnown
  desc: Set remaining unknown words to Well Known, mark page as read, go to next page

Translate:
- hotkey: hotkey_TranslateSentence
  desc: Translate the sentence of the current word
- hotkey: hotkey_TranslatePara
  desc: Translate the paragraph of the current word
- hotkey: hotkey_TranslatePage
  desc: Translate the full page

Update status:
- hotkey: hotkey_Status1
  desc: Set status to 1
- hotkey: hotkey_Status2
  desc: Set status to 2
- hotkey: hotkey_Status3
  desc: Set status to 3
- hotkey: hotkey_Status4
  desc: Set status to 4
- hotkey: hotkey_Status5
  desc: Set status to 5
- hotkey: hotkey_StatusIgnore
  desc: Set status to Ignore
- hotkey: hotkey_StatusWellKnown
  desc: Set status to Well Known
- hotkey: hotkey_StatusUp
  desc: Bump the status up by 1
- hotkey: hotkey_StatusDown
  desc: Bump that status down by 1
- hotkey: hotkey_DeleteTerm
  desc: Delete term (set status to Unknown)
"""


# Initial hotkey values
#
# Only *some* of the hotkeys have default values assigned, as these
# hotkeys were initially set in the early releases of Lute.
#
# Any new hotkeys added *MUST NOT* have defaults assigned, as users
# may have already setup their hotkeys, and we can't assume that a
# given key combination is free:
_initial_values = {
    "hotkey_StartHover": "Escape",
    "hotkey_PrevWord": "ArrowLeft",
    "hotkey_NextWord": "ArrowRight",
    "hotkey_Status1": "Digit1",
    "hotkey_Status2": "Digit2",
    "hotkey_Status3": "Digit3",
    "hotkey_Status4": "Digit4",
    "hotkey_Status5": "Digit5",
    "hotkey_StatusIgnore": "KeyI",
    "hotkey_StatusWellKnown": "KeyW",
    "hotkey_StatusUp": "ArrowUp",
    "hotkey_StatusDown": "ArrowDown",
    "hotkey_TranslateSentence": "KeyT",
    "hotkey_TranslatePara": "shift+KeyT",
    "hotkey_CopySentence": "KeyC",
    "hotkey_CopyPara": "shift+KeyC",
    "hotkey_Bookmark": "KeyB",
    "hotkey_NextTheme": "KeyM",
    "hotkey_ToggleHighlight": "KeyH",
    "hotkey_ToggleFocus": "KeyF",
    "hotkey_SaveTerm": "ctrl+Enter",
}


def initial_hotkey_defaults():
    """
    Get initial hotkeys and defaults (or empty string).
    Used for db initialization.
    """
    y = yaml.safe_load(_ALL_HOTKEY_DATA)
    hks = []
    for _, keys in y.items():
        hks.extend([k["hotkey"] for k in keys])
    ret = {h: _initial_values.get(h, "") for h in hks}
    return ret


def categorized_hotkeys():
    "Hotkeys by category.  Used by routes."
    y = yaml.safe_load(_ALL_HOTKEY_DATA)
    ordered_keys = [
        "Navigation",
        "Update status",
        "Paging",
        "Translate",
        "Copy",
        "Misc",
    ]
    if set(ordered_keys) != set(y.keys()):
        raise RuntimeError("ordered_keys doesn't match expected")
    ret = {k: [h["hotkey"] for h in y[k]] for k in ordered_keys}
    return ret


def hotkey_descriptions():
    """
    Get hotkeys and descriptions.  Used by routes.
    """
    y = yaml.safe_load(_ALL_HOTKEY_DATA)
    return {key["hotkey"]: key["desc"] for group in y.values() for key in group}


if __name__ == "__main__":
    print("---")
    pprint.pprint(categorized_hotkeys())
    print("---")
    pprint.pprint(initial_hotkey_defaults())
    print("---")
    pprint.pprint(hotkey_descriptions())



"""
Field-to-value tests.
"""

from unittest.mock import Mock
import pytest

from lute.ankiexport.exceptions import AnkiExportConfigurationError
from lute.ankiexport.field_mapping import (
    get_values_and_media_mapping,
    validate_mapping,
    get_fields_and_final_values,
    SentenceLookup,
)

# pylint: disable=missing-function-docstring


@pytest.mark.parametrize(
    "mapping,msg",
    [
        ({"a": "{ x }"}, 'Invalid field mapping "x"'),
        ({"a": "{ id }", "b": "{ x }"}, 'Invalid field mapping "x"'),
    ],
)
def test_validate_mapping_throws_if_bad_mapping_string(mapping, msg):
    with pytest.raises(AnkiExportConfigurationError, match=msg):
        validate_mapping(mapping)


@pytest.mark.parametrize(
    "mapping,msg",
    [
        ({"a": "{ id }"}, "ok"),
        ({"a": "{ id }", "b": "{ id }"}, "same value twice ok"),
        ({"a": "{ id }", "b": "{ term }"}, "different fields"),
    ],
)
def test_validate_mapping_does_not_throw_if_ok(mapping, msg):
    validate_mapping(mapping)
    assert True, msg


@pytest.fixture(name="term")
def fixture_term():
    zws = "\u200B"
    term = Mock()
    term.id = 1
    term.text = f"test{zws} {zws}term"
    term.language.name = "English"
    term.language.id = 42
    term.get_current_image.return_value = "image.jpg"
    term.parents = []
    term.term_tags = [Mock(text="noun"), Mock(text="verb")]
    term.translation = f"example{zws} {zws}translation"
    return term


def test_basic_replacements(term):
    sentence_lookup = Mock()
    mapping = {
        "id": "{ id }",
        "term": "{ term }",
        "language": "{ language }",
        "translation": "{ translation }",
    }
    values, media = get_values_and_media_mapping(term, sentence_lookup, mapping)

    expected = {
        "id": 1,
        "term": "test term",
        "parents": "",
        "tags": "noun, verb",
        "language": "English",
        "translation": "example translation",
    }
    assert values == expected, "mappings"
    assert len(media) == 0


def test_tag_replacements(term):
    sentence_lookup = Mock()
    mapping = {"tags": "{ tags }"}
    values, media = get_values_and_media_mapping(term, sentence_lookup, mapping)

    assert set(values["tags"].split(", ")) == {"noun", "verb"}
    assert len(media) == 0


def test_filtered_tag_replacements(term):
    sentence_lookup = Mock()
    mapping = {"mytags": '{ tags:["noun"] }'}
    values, media = get_values_and_media_mapping(term, sentence_lookup, mapping)
    assert set(values['tags:["noun"]'].split(", ")) == {"noun"}
    assert len(media) == 0


def test_image_handling(term):
    sentence_lookup = Mock()
    mapping = {"image": "{ image }"}

    values, media = get_values_and_media_mapping(term, sentence_lookup, mapping)

    assert media == {"LUTE_TERM_1.jpg": "/userimages/42/image.jpg"}, "one image"
    assert '<img src="LUTE_TERM_1.jpg">' in values["image"]


def test_sentence_handling(term):
    zws = "\u200B"
    sentence_lookup = Mock()
    sentence_lookup.get_sentence_for_term.return_value = f"Example{zws} {zws}sentence."
    mapping = {"sentence": "{ sentence }"}

    values, media = get_values_and_media_mapping(term, sentence_lookup, mapping)

    assert values["sentence"] == "Example sentence."
    assert len(media) == 0


def test_get_fields_and_final_values_smoke_test():
    mapping = {
        "a": "{ id }",
        "b": "{ term }",
    }
    replacements = {"id": 42, "term": "rabbit"}
    actual = get_fields_and_final_values(mapping, replacements)
    assert actual == {"a": "42", "b": "rabbit"}


def test_empty_fields_not_posted():
    mapping = {
        "a": "{ id }",
        "b": "{ term }",
    }
    replacements = {"id": 42, "term": ""}
    actual = get_fields_and_final_values(mapping, replacements)
    assert actual == {"a": "42"}


def test_sentence_lookup_finds_sentence_in_supplied_dict_or_does_db_call():
    refsrepo = Mock()
    refsrepo.find_references_by_id.return_value = {"term": [Mock(sentence="Db lookup")]}
    fixed_sentences = {"42": "Hello"}
    lookup = SentenceLookup(fixed_sentences, refsrepo)
    assert lookup.get_sentence_for_term("42") == "Hello", "looks up"
    assert lookup.get_sentence_for_term(42) == "Hello", "int ok, still finds"
    assert lookup.get_sentence_for_term(99) == "Db lookup", "falls back to db lookup"
    assert lookup.get_sentence_for_term("99") == "Db lookup", "falls back to db lookup"



"""
Term Repository tests, with syncing status to parent.
"""

import pytest
from lute.term.model import Repository
from lute.db import db
from tests.dbasserts import assert_sql_result
from tests.utils import add_terms

# pylint: disable=unbalanced-tuple-unpacking, missing-function-docstring,unused-argument


@pytest.fixture(name="repo")
def fixture_repo(app_context):
    return Repository(db.session)


@pytest.fixture(name="t")
def fixture_term(repo, spanish):
    [dt] = add_terms(spanish, ["T"])
    dt.status = 1
    db.session.add(dt)
    db.session.commit()
    yield repo.load(dt.id)


@pytest.fixture(name="p")
def fixture_parent(repo, spanish):
    [dp] = add_terms(spanish, ["P"])
    dp.status = 4
    db.session.add(dp)
    db.session.commit()
    yield repo.load(dp.id)


def assert_statuses(term, expected, msg):
    repo = Repository(db.session)
    repo.add(term)
    repo.commit()
    sql = "select WoText, WoStatus from words order by WoText"
    assert_sql_result(sql, expected, msg)


def test_save_synced_child_with_set_status_updates_parent_status(t, p):
    t.parents = ["P"]
    t.sync_status = True
    t.status = 3
    assert_statuses(t, ["P; 3", "T; 3"], "both updated")

    t.sync_status = False
    t.status = 1
    assert_statuses(t, ["P; 3", "T; 1"], "P not updated")


def test_save_child_with_no_set_status_inherits_parent_status_if_sync(t, p):
    assert_statuses(t, ["P; 4", "T; 1"], "initial state")
    t.parents = ["P"]
    t.sync_status = True
    assert_statuses(t, ["P; 4", "T; 4"], "initial save")


def test_new_parent_gets_child_status_if_sync(t):
    t.parents = ["NEW"]
    t.sync_status = True
    assert_statuses(t, ["NEW; 1", "T; 1"], "initial save")


def test_child_status_changes_dont_affect_others_if_no_sync(t, p):
    t.status = 3
    assert_statuses(t, ["P; 4", "T; 3"], "no change to other")


def test_remove_parent_breaks_sync(t, p):
    t.parents = ["P"]
    t.sync_status = True
    t.status = 3
    assert_statuses(t, ["P; 3", "T; 3"], "both updated")

    t.parents = []
    t.status = 1
    assert_statuses(t, ["P; 3", "T; 1"], "P not updated, no sync")


def test_cant_sync_multiple_parents(t, p):
    t.parents = ["P", "X"]
    t.sync_status = True
    t.status = 2
    assert_statuses(
        t, ["P; 4", "T; 2", "X; 2"], "P no change, X.status defaults to T's"
    )

    t.status = 3
    assert_statuses(t, ["P; 4", "T; 3", "X; 2"], "X not sync'd, multiple parents")

    t.parents = ["X"]
    assert_statuses(t, ["P; 4", "T; 3", "X; 3"], "X sync'd, single parent")



"""
Getting and saving bing image search results.
"""

import os
import datetime
import hashlib
import re
import urllib.request
from flask import (
    Blueprint,
    request,
    Response,
    render_template,
    jsonify,
    current_app,
    url_for,
)


bp = Blueprint("bing", __name__, url_prefix="/bing")


@bp.route(
    "/search_page/<int:langid>/<string:text>/<string:searchstring>", methods=["GET"]
)
def bing_search_page(langid, text, searchstring):
    """
    Load initial empty search page, passing real URL for subsequent ajax call to get images.

    Sometimes Bing image searches block or fail, so providing the initial empty search page
    lets the user know work is in progress.  The user can therefore interact with the page
    immediately. The template for this route then makes an ajax call to the "bing_search()"
    method below which actually does the search.
    """

    # Create URL for bing_search and pass into template.
    search_url = url_for(
        "bing.bing_search", langid=langid, text=text, searchstring=searchstring
    )

    return render_template(
        "imagesearch/index.html", langid=langid, text=text, search_url=search_url
    )


@bp.route("/search/<int:langid>/<string:text>/<string:searchstring>", methods=["GET"])
def bing_search(langid, text, searchstring):
    "Do an image search."

    # Searching for images slows acceptance tests.  If NO_BING_IMAGES
    # environment setting, don't do a search.
    if "NO_BING_IMAGES" in os.environ:
        return render_template(
            "imagesearch/index.html", langid=langid, text=text, images=[]
        )

    # dump("searching for " + text + " in " + language.getLgName())
    search = urllib.parse.quote(text)
    params = searchstring.replace("[LUTE]", search)
    params = params.replace("###", search)  # TODO remove_old_###_placeholder: remove
    url = "https://www.bing.com/images/search?" + params
    content = ""
    error_msg = ""
    try:
        with urllib.request.urlopen(url) as s:
            content = s.read().decode("utf-8")
    except urllib.error.URLError as e:
        content = ""
        error_msg = str(e.reason)
    except Exception as e:  # pylint: disable=broad-exception-caught
        content = ""
        error_msg = str(e)

    # Sample data returned by bing image search:
    # <img class="mimg vimgld" ... data-src="https:// ...">
    # or
    # <img class="mimg rms_img" ... src="https://tse4.mm.bing ..." >

    def is_search_img(img):
        return not ('src="/' in img) and ("rms_img" in img or "vimgld" in img)

    def build_struct(image):
        src = "missing"
        normalized_source = image.replace("data-src=", "src=")
        m = re.search(r'src="(.*?)"', normalized_source)
        if m:
            src = m.group(1)
        return {"html": image, "src": src}

    raw_images = list(re.findall(r"(<img .*?>)", content, re.I))

    images = [build_struct(i) for i in raw_images if is_search_img(i)]

    # Reduce image load count so we don't kill subpage loading.
    # Also bing seems to throttle images if the count is higher (??).
    images = images[:25]

    ret = {
        "langid": langid,
        "text": text,
        "images": images,
        "error_message": error_msg,
    }
    return jsonify(ret)


def _get_dir_and_filename(langid, text):
    "Make a directory if needed, return [dir, filename]"
    datapath = current_app.config["DATAPATH"]
    image_dir = os.path.join(datapath, "userimages", langid)
    if not os.path.exists(image_dir):
        os.makedirs(image_dir)

    now = datetime.datetime.now()
    timestamp = now.strftime("%Y%m%d_%H%M%S%f")[:-3]
    hash_part = hashlib.md5(text.encode()).hexdigest()[:8]
    filename = f"{timestamp}_{hash_part}.jpeg"
    return [image_dir, filename]


@bp.route("/save", methods=["POST"])
def bing_save():
    """
    Save the posted image data to DATAPATH/userimages,
    returning the filename.
    """
    src = request.form["src"]
    text = request.form["text"]
    langid = request.form["langid"]

    imgdir, filename = _get_dir_and_filename(langid, text)
    destfile = os.path.join(imgdir, filename)
    with urllib.request.urlopen(src) as response, open(destfile, "wb") as out_file:
        out_file.write(response.read())

    ret = {
        "url": f"/userimages/{langid}/{filename}",
        "filename": filename,
    }
    return jsonify(ret)


@bp.route("/manual_image_post", methods=["POST"])
def manual_image_post():
    """
    For manual posts of images (not bing image clicks).
    Save the posted image data to DATAPATH/userimages,
    returning the filename.
    """
    text = request.form["text"]
    langid = request.form["langid"]

    if "manual_image_file" not in request.files:
        return Response("No file part in request", status=400)

    f = request.files["manual_image_file"]
    if f.filename == "":
        return Response("No selected file", status=400)

    imgdir, filename = _get_dir_and_filename(langid, text)
    destfile = os.path.join(imgdir, filename)
    f.save(destfile)

    ret = {
        "url": f"/userimages/{langid}/{filename}",
        "filename": filename,
    }
    return jsonify(ret)



"""
Lute test client.

Convenience wrapper around the browser.

IMPORTANT: on start, clears db, and disables backup.

Adds the home url to visited urls,
provides common functions like make_book.

Also calls the dev_api to set state, etc.

This module is "registered" to pytest in ./__init__.py
to get nicer assertion details.
"""

import time
import json
import requests
from selenium.webdriver.common.keys import Keys
from selenium.webdriver.common.action_chains import ActionChains


class LuteTestClient:  # pylint: disable=too-many-public-methods
    """
    The client!
    """

    ################################3
    # Setup

    def __init__(self, b, home):
        self.browser = b
        self.home = home
        self.visit("dev_api/wipe_db")
        self.visit("dev_api/load_demo_languages")
        self.visit("dev_api/disable_backup")

        response = requests.get(f"{home}/dev_api/language_ids", timeout=5)
        self.language_ids = response.json()

        self.start = time.perf_counter()
        self.last_step = self.start

    def load_demo_stories(self):
        "Load the demo stories."
        self.visit("dev_api/load_demo_stories")
        self.visit("/")
        self.clear_book_filter()

    def change_parser_registry_key(self, key, replacement):
        """
        Change a parser registry key to a replacement,
        effectively disabling that parser.
        """
        self.visit(f"dev_api/disable_parser/{key}/{replacement}")

    ################################3
    # Browsing

    def visit(self, suburl):
        "Visit a sub url under the base."
        if suburl.startswith("/"):
            suburl = suburl[1:]
        url = f"{self.home}/{suburl}"
        # print(f'visiting: {url}')
        self.browser.visit(url)

    def index(self):
        "Go to home page."
        self.browser.visit("")

    def clear_book_filter(self):
        "Clear all state.  Normally state is saved."
        self.browser.execute_script("clear_datatable_state()")
        time.sleep(0.1)

    def click_link(self, linktext):
        self.browser.links.find_by_text(linktext).click()

    ################################3
    # Languages

    def edit_language(self, langname, updates=None):
        """
        Edit a language.
        """
        self.visit("/")
        self.browser.find_by_css("#menu_settings").mouse_over()
        self.browser.find_by_id("lang_index").first.click()
        # WEIRD: find_by_text(langname) doesn't work ...
        self.browser.links.find_by_partial_text(langname).click()
        assert f"Edit {langname}" in self.browser.html
        updates = updates or {}
        for k, v in updates.items():
            if k == "exceptions_split_sentences":
                self.browser.find_by_css(f"#{k}").fill(v)
            else:
                raise RuntimeError(f"unhandled key {k}")
        self.browser.find_by_css("#submit").first.click()

    ################################3
    # Books

    def make_book(self, title, text, langname):
        "Create a book with title, text, and languagename."
        self.visit("book/new")
        self.browser.fill("text", text)
        self.browser.find_by_css("#title").fill(title)
        self.browser.select("language_id", int(self.language_ids[langname]))
        self.browser.find_by_css("#save").first.click()

    def make_book_from_file(self, title, filename, langname):
        "Create a book with title, content from filename, and languagename."
        self.visit("book/new")
        self.browser.attach_file("textfile", filename)
        self.browser.find_by_css("#title").fill(title)
        self.browser.select("language_id", int(self.language_ids[langname]))
        self.browser.find_by_css("#save").first.click()

    def make_book_from_url(self, url, langname):
        "Create a book with title, content from url, and languagename."
        self.visit("book/import_webpage")
        self.browser.find_by_css("#importurl").fill(url)
        self.browser.find_by_css("#import").first.click()
        time.sleep(0.1)  # hack
        self.browser.select("language_id", int(self.language_ids[langname]))
        self.browser.find_by_css("#save").first.click()

    def get_book_table_content(self):
        "Get book table content."
        css = "#booktable tbody tr"

        # Skip the last two columns:
        # - "last opened" date is a hassle to check
        # - "actions" is just "..."
        def _to_string(row):
            tds = row.find_by_css("td")
            rowtext = [td.text.strip() for td in tds]
            ret = "; ".join(rowtext[:-2]).strip()
            # print(ret, flush=True)
            return ret

        rows = list(self.browser.find_by_css(css))
        return "\n".join([_to_string(row) for row in rows])

    def get_book_page_start_dates(self):
        "get content from sql check"
        sql = """select bktitle, txorder
        from books
        inner join texts on txbkid = bkid
        where txstartdate is not null
        order by bktitle, txorder"""
        response = requests.get(f"{self.home}/dev_api/sqlresult/{sql}", timeout=1)
        ret = "\n".join(json.loads(response.text))
        if ret == "":
            ret = "-"
        return ret

    def get_book_page_read_dates(self):
        "get content from sql check"
        sql = """select bktitle, txorder
        from books
        inner join texts on txbkid = bkid
        where txreaddate is not null
        order by bktitle, txorder"""
        response = requests.get(f"{self.home}/dev_api/sqlresult/{sql}", timeout=1)
        ret = "\n".join(json.loads(response.text))
        if ret == "":
            ret = "-"
        return ret

    def set_txstartdate_to_null(self):
        "hack back end to keep test data sane."
        sql = "update texts set txstartdate = null"
        response = requests.get(f"{self.home}/dev_api/execsql/{sql}", timeout=1)
        return response.text

    ################################
    # Terms

    def _fill_tagify_field(self, b, fieldid, text):
        "Fill in field in browser b with text."
        xpath = [
            # input w/ id
            f'//input[@id="{fieldid}"]',
            # <tags> before it.
            "/preceding-sibling::tags",
            # <span> within the <tags> with class.
            '/span[@class="tagify__input"]',
        ]
        xpath = "".join(xpath)

        # Sometimes test runs couldn't find the parent
        # tagify input, so hacky loop to get it and retry.
        span = None
        attempts = 0
        while span is None and attempts < 10:
            time.sleep(0.2)  # seconds
            attempts += 1
            span = b.find_by_xpath(xpath)
        if span is None:
            raise RuntimeError(f"unable to find {fieldid}")

        span.type(text, slowly=False)
        span.type(Keys.RETURN)
        time.sleep(0.3)  # seconds

    def _fill_term_form(self, b, updates):
        "Fill in the term form."
        for k, v in updates.items():
            if k == "language_id":
                b.select("language_id", v)
            elif k == "status":
                # This line didn't work:
                # iframe.choose('status', updates['status'])
                s = updates["status"]
                xp = "".join(
                    [
                        "//input[@type='radio'][@name='status']",
                        f"[@value='{s}']",
                        "/following-sibling::label",
                    ]
                )
                labels = b.find_by_xpath(xp)
                assert len(labels) == 1, "have matching radio button"
                label = labels[0]
                label.click()
            elif k in ("translation", "text", "romanization"):
                b.find_by_css(f"#{k}").fill(v)
            elif k in ("pronunciation"):
                b.find_by_css("#romanization").fill(v)
            elif k == "parents":
                for p in updates["parents"]:
                    self._fill_tagify_field(b, "parentslist", p)
            elif k == "sync_status":
                if v:
                    b.check("sync_status")
                else:
                    b.uncheck("sync_status")
            else:
                raise RuntimeError(f"unhandled key {k}")

    def _fill_bulk_term_edit_form(self, b, updates):
        "Fill in the term bulk edit form."
        for k, v in updates.items():
            if k == "remove parents":
                if v:
                    b.check("remove_parents")
                else:
                    b.uncheck("remove_parents")
            elif k == "parent":
                self._fill_tagify_field(b, "txtSetParent", v)
            elif k == "change status":
                if v:
                    b.check("change_status")
                else:
                    b.uncheck("change_status")
            elif k == "status":
                # This line didn't work:
                # iframe.choose('status', updates['status'])
                s = updates["status"]
                xp = "".join(
                    [
                        "//input[@type='radio'][@name='status']",
                        f"[@value='{s}']",
                        "/following-sibling::label",
                    ]
                )
                labels = b.find_by_xpath(xp)
                assert len(labels) == 1, "have matching radio button"
                label = labels[0]
                label.click()
            elif k in ("add tags", "remove tags"):
                fields = {"add tags": "txtAddTags", "remove tags": "txtRemoveTags"}
                for tag in updates[k].split(", "):
                    self._fill_tagify_field(b, fields[k], tag)
            else:
                raise RuntimeError(f"unhandled key {k}")

    def make_term(self, lang, updates):
        "Create a new term."
        # Sometimes this failed during the unsupported_parser.feature, not sure why.
        # Likely something silly, don't care, so will bypass the screen controls.
        # I am sure this will bite me later.
        # TODO fix_nav: figure out why occasionally got ElementNotInteractableException
        # self.visit("/")
        # self.browser.find_by_css("#menu_terms").mouse_over()
        # self.browser.find_by_id("term_index").first.click()
        # self.browser.find_by_css("#term_actions").mouse_over()
        # self.click_link("Create new")
        self.visit("/term/new")
        assert "New Term" in self.browser.html

        updates["language_id"] = self.language_ids[lang]
        b = self.browser
        self._fill_term_form(b, updates)
        b.find_by_css("#btnsubmit").first.click()

    def get_term_table_content(self):
        "Get term table content."
        self.visit("/")
        self.browser.find_by_css("#menu_terms").mouse_over()
        self.browser.find_by_id("term_index").first.click()

        # clear any filters!
        self.browser.find_by_id("showHideFilters").first.click()

        # The last column of the table is the "date added", but that's
        # a hassle to check, so ignore it.
        def _to_string(row):
            tds = row.find_by_css("td")
            rowtext = [td.text.strip() for td in tds]
            check = "; ".join(rowtext).strip()
            if check == "No data available in table":
                return check

            rowtext = [""]  # first field is empty checkbox
            rowtext.append(tds[1].text.strip())  # term

            parenttags = tds[2].text.strip()
            parenttags = ", ".join(parenttags.split("\n"))
            rowtext.append(parenttags)

            rowtext.append(tds[3].text.strip())  # translation
            rowtext.append(tds[6].text.strip())  # language

            termtags = tds[4].text.strip()
            termtags = ", ".join(termtags.split("\n"))
            rowtext.append(termtags)

            select_element = row.find_by_css("select")
            selected_value = select_element.value
            selected_option = select_element.find_by_css(
                f'option[value="{selected_value}"]'
            )
            rowtext.append(selected_option.text)  # status select
            return "; ".join(rowtext).strip()

        css = "#termtable tbody tr"
        rows = list(self.browser.find_by_css(css))
        rowstring = [_to_string(row) for row in rows]
        return "\n".join(rowstring)

    ################################3
    # Reading/rendering

    def displayed_text(self):
        "Return the TextItems, with '/' at token boundaries."
        elements = self.browser.find_by_xpath('//span[contains(@class, "textitem")]')

        def _to_string(t):
            "Create string for token, eg 'cat (2)'."

            # Note that selenium's t.text accessor strips leading/trailing whitespace,
            # so if a span contains " ", t.text returns "".  We need the actual
            # inner html.
            # pylint: disable=protected-access
            inner_html = t._element.get_attribute("innerHTML")
            zws = "\u200B"
            inner_html = inner_html.replace(zws, "")

            status = [
                c.replace("status", "")
                for c in t["class"].split(" ")
                if c.startswith("status") and c != "status0"
            ]
            if len(status) == 0:
                return inner_html
            assert len(status) == 1, f"should only have 1 status on {inner_html}"
            return f"{inner_html} ({status[0]})"

        etext = [_to_string(e) for e in elements]
        ret = "/".join(etext)
        if ret.endswith("/"):
            ret = ret[:-1]
        return ret

    ################################3
    # Reading, term actions

    def _get_element_for_word(self, word):
        "Helper, get the element."
        # print('getting ' + word)
        elements = self.browser.find_by_xpath('//span[contains(@class, "textitem")]')
        es = [e for e in elements if e.text == word]
        assert len(es) > 0, f"match for {word}"
        return es[0]

    def click_word(self, word):
        "Click a word in the reading frame."
        self._get_element_for_word(word).click()

    def shift_click_words(self, words):
        "Shift-click words."
        # https://stackoverflow.com/questions/27775759/
        #   send-keys-control-click-in-selenium-with-python-bindings
        # pylint: disable=protected-access
        els = [self._get_element_for_word(w)._element for w in words]
        ac = ActionChains(self.browser.driver).key_down(Keys.SHIFT)
        for e in els:
            ac = ac.click(e)
        ac = ac.key_up(Keys.SHIFT)
        ac.perform()

    def shift_drag(self, fromword, toword):
        "Shift-drag over words."
        # https://stackoverflow.com/questions/27775759/
        #   send-keys-control-click-in-selenium-with-python-bindings
        # pylint: disable=protected-access
        [fromel, toel] = [
            self._get_element_for_word(w)._element for w in [fromword, toword]
        ]
        actions = ActionChains(self.browser.driver)
        actions.key_down(Keys.SHIFT)
        actions.click_and_hold(fromel)
        actions.move_to_element(toel)
        actions.key_up(Keys.SHIFT)
        actions.release()
        actions.perform()

    def drag(self, fromword, toword):
        "drag over words."
        # https://stackoverflow.com/questions/27775759/
        #   send-keys-control-click-in-selenium-with-python-bindings
        # pylint: disable=protected-access
        [fromel, toel] = [
            self._get_element_for_word(w)._element for w in [fromword, toword]
        ]
        actions = ActionChains(self.browser.driver)
        actions.click_and_hold(fromel)
        actions.move_to_element(toel)
        actions.release()
        actions.perform()

    def _refresh_browser(self):
        """
        Term actions (edits, hotkeys) cause updated content to be ajaxed in.
        For the splinter browser to be aware of it, the browser has to be
        reloaded, but calling a self.browser.reload() has other side effects
        (sets the page start date, etc).

        The below weird js hack causes the browser to be updated,
        and then the js events have to be reattached too.
        """
        # self.browser.reload()
        # ??? ChatGPT suggested:
        time.sleep(0.5)  # Hack for ci.
        self.browser.execute_script(
            """
            // Trigger re-render of the entire body
            var body = document.querySelector('body');
            var content = body.innerHTML;
            body.innerHTML = '';
            body.innerHTML = content;

            // Re-attach text interactions.
            window.prepareTextInteractions();
            """
        )
        time.sleep(0.5)  # Hack for ci.

    def fill_reading_bulk_edit_form(self, updates=None):
        """
        Click a word in the reading frame, fill in the term form iframe.
        """
        updates = updates or {}
        should_refresh = False
        with self.browser.get_iframe("wordframe") as iframe:
            time.sleep(0.4)  # Hack, test failing.
            self._fill_bulk_term_edit_form(iframe, updates)
            time.sleep(0.4)  # Hack, test failing.
            iframe.find_by_css("#btnsubmit").first.click()
            time.sleep(0.4)  # Hack, test failing.

            # Only refresh the reading frame if everything was ok.
            # Some submits will fail due to validation errors,
            # and we want to look at them.
            if "updated" in iframe.html:
                should_refresh = True

        # Have to refresh the content to query the dom.
        if should_refresh:
            self._refresh_browser()

    def hack_set_hotkey(self, hotkey, value):
        "Hack set hotkey directly through dev api.  Trashy."
        sql = f"""update settings
        set StValue='{value}' where StKey='{hotkey}'"""
        requests.get(f"{self.home}/dev_api/execsql/{sql}", timeout=1)
        # NOTE! Hacking is dumb, it bypassing the global state which is rendered in JS.
        # Have to visit and save settings to re-set the JS values that will be rendered.
        # Big time waste finding this out.
        self.visit("settings/shortcuts")
        self.browser.find_by_css("#btnSubmit").first.click()

    def press_hotkey(self, hotkey):
        "Send a hotkey."
        key_to_code_map = {
            "escape": "Escape",
            "1": "Digit1",
            "2": "Digit2",
            "3": "Digit3",
            "4": "Digit4",
            "5": "Digit5",
            "arrowdown": "ArrowDown",
            "arrowup": "ArrowUp",
            "h": "KeyH",
            "i": "KeyI",
            "m": "KeyM",
            "w": "KeyW",
            # Manually added.
            "8": "Digit8",
            "9": "Digit9",
        }
        if hotkey.lower() not in key_to_code_map:
            raise RuntimeError(f"Missing {hotkey} in acceptance test map")
        event_parts = [
            "type: 'keydown'",
            f"code: '{key_to_code_map[hotkey.lower()]}'",
        ]
        if hotkey != hotkey.lower():
            event_parts.append("shiftKey: true")
        script = f"""jQuery.event.trigger({{
          {', '.join(event_parts)}
        }});"""
        # print(script, flush=True)
        # pylint: disable=protected-access
        el = self.browser.find_by_id("thetext")
        self.browser.execute_script(script, el._element)
        time.sleep(0.2)  # Or it's too fast.
        # print(script)
        # Have to refresh the content to query the dom.
        self._refresh_browser()

    def click_word_fill_form(self, word, updates=None):
        """
        Click a word in the reading frame, fill in the term form iframe.
        """
        self.click_word(word)
        updates = updates or {}

        should_refresh = False
        with self.browser.get_iframe("wordframe") as iframe:
            time.sleep(0.4)  # Hack, test failing.
            self._fill_term_form(iframe, updates)
            time.sleep(0.4)  # Hack, test failing.
            iframe.find_by_css("#btnsubmit").first.click()
            time.sleep(0.4)  # Hack, test failing.

            # Only refresh the reading frame if everything was ok.
            # Some submits will fail due to validation errors,
            # and we want to look at them.
            if "updated" in iframe.html:
                should_refresh = True

        # Have to refresh the content to query the dom.
        if should_refresh:
            self._refresh_browser()

    ################################3
    # Misc.

    def elapsed(self, step):
        """
        Helper method for sorting out slowness.

        For the step, gives elapsed time since start of
        the LuteBrowser, and since the last recorded step.

        To see this data, run the acc. tests with '-s', eg:

        pytest tests/acceptance/test_smoke.py --port=5001 -s
        """
        now = time.perf_counter()
        since_start = now - self.start
        print(step)
        print(f"total elapsed: {since_start}")
        print(f"since last:    {now - self.last_step}")
        self.last_step = now

    def sleep(self, seconds):
        "Nap."
        time.sleep(seconds)

    def get_temp_file_content(self, filename):
        "Get book table content."
        response = requests.get(
            f"{self.home}/dev_api/temp_file_content/{filename}", timeout=1
        )
        return response.text



"""
token_group_iterator tests.
"""

from lute.book.model import token_group_generator
from lute.parse.space_delimited_parser import SpaceDelimitedParser


def toks_to_string(tokens):
    ret = "".join([t.token for t in tokens])
    return ret.replace("¶", "\n").strip()


def test_paragraph_scenarios(english):
    """
    Given a string and the max token count,
    generator should return expected groups.
    """

    def scenario(s, threshold, expected_groups):
        parser = SpaceDelimitedParser()
        tokens = parser.get_parsed_tokens(s, english)
        # pylint: disable=unnecessary-comprehension
        groups = [g for g in token_group_generator(tokens, "paragraphs", threshold)]
        print(groups, flush=True)
        gs = [toks_to_string(g) for g in groups]
        expected = "||".join(expected_groups)
        assert "||".join(gs) == expected, f"groups for size {threshold}"

    scenario("", 100, [""])

    scenario("Here is a dog. Here is a cat.", 100, ["Here is a dog. Here is a cat."])
    scenario("Here is a dog. Here is a cat.", 3, ["Here is a dog. Here is a cat."])
    scenario("Here is a dog. Here is a cat", 6, ["Here is a dog. Here is a cat"])
    scenario("Here is a dog Here is a cat", 6, ["Here is a dog Here is a cat"])

    p1 = "Here is a dog."
    p2 = "And a cat."
    src = "\n".join([p1, p2])
    scenario(src, 100, [src])
    scenario(src, 5, [src])
    scenario(src, 3, [p1, p2])
    scenario(src.replace("\n", "\n\n\n"), 3, [p1, p2])

    p1 = "Here is a dog."
    p2 = "Here is a cat. Much more info here, long paragraph."
    p3 = "Last stuff."
    src = "\n".join([p1, p2, p3])
    scenario(src, 100, [src])
    scenario(src, 5, [f"{p1}\n{p2}", p3])
    scenario(src, 3, [p1, p2, p3])

    src = "\n" + "\n\n\n".join([p1, p2, p3]) + "\n"
    scenario(src, 5, [f"{p1}\n\n\n{p2}", p3])


def test_sentence_scenarios(english):
    """
    Given a string and the max token count,
    generator should return expected groups.
    """

    def scenario(s, threshold, expected_groups):
        parser = SpaceDelimitedParser()
        tokens = parser.get_parsed_tokens(s, english)
        # pylint: disable=unnecessary-comprehension
        groups = [g for g in token_group_generator(tokens, "sentences", threshold)]
        gs = [toks_to_string(g) for g in groups]
        expected = "||".join(expected_groups)
        assert "||".join(gs) == expected, f"groups for size {threshold}"

    scenario("", 100, [""])

    text = "Here is a dog. Here is a cat."

    scenario(text, 100, ["Here is a dog. Here is a cat."])

    scenario(text, 3, ["Here is a dog.", "Here is a cat."])

    scenario(text, 6, ["Here is a dog. Here is a cat."])

    # No period at the end.
    scenario("Here is a dog. Here is a cat", 6, ["Here is a dog. Here is a cat"])

    # No period at all.
    scenario("Here is a dog Here is a cat", 6, ["Here is a dog Here is a cat"])

    s1 = "Here is a dog."
    s2 = "Here is a cat."
    s3 = "Here is a thing."
    src = " ".join([s1, s2, s3])
    scenario(src, 10, [src])
    scenario(src, 7, [f"{s1} {s2}", s3])
    scenario(src, 3, [s1, s2, s3])



"""
Term Repository tests.

Tests lute.term.model.Term *domain* objects being saved
and retrieved from DB.
"""

from datetime import datetime
import pytest

from lute.db import db
from lute.term.model import Term, Repository, ReferencesRepository
from tests.dbasserts import assert_record_count_equals
from tests.utils import add_terms, make_text


@pytest.fixture(name="repo")
def fixture_repo():
    return Repository(db.session)


@pytest.fixture(name="refsrepo")
def fixture_refs_repo():
    return ReferencesRepository(db.session)


## Sentences


def full_refs_to_string(refs):
    "Convert refs to strings for simpler testing."

    def to_string(r):
        return f"{r.title}, {r.sentence}"

    def refs_to_string(refs_array):
        ret = [to_string(r) for r in refs_array]
        ret.sort()
        return ret

    def parent_refs_to_string(prefs):
        ret = []
        for p in prefs:
            ret.append({"term": p["term"], "refs": refs_to_string(p["refs"])})
        return ret

    return {
        "term": refs_to_string(refs["term"]),
        "children": refs_to_string(refs["children"]),
        "parents": parent_refs_to_string(refs["parents"]),
    }


@pytest.mark.sentences
def test_get_all_references(spanish, repo, refsrepo):
    "Check references with parents and children."
    text = make_text(
        "hola", "Tengo un gato.  Ella tiene un perro.  No quiero tener nada.", spanish
    )
    archtext = make_text("luego", "Tengo un coche.", spanish)
    archtext.book.archived = True
    db.session.add(archtext.book)

    for t in [text, archtext]:
        t.read_date = datetime.now()
        db.session.add(t)

    db.session.commit()

    tengo = Term()
    tengo.language_id = spanish.id
    tengo.text = "tengo"
    tengo.parents = ["tener"]
    repo.add(tengo)

    tiene = Term()
    tiene.language_id = spanish.id
    tiene.text = "tiene"
    tiene.parents = ["tener"]
    repo.add(tiene)

    repo.commit()

    refs = refsrepo.find_references(tengo)
    assert full_refs_to_string(refs) == {
        "term": [
            "hola (1/1), <b>Tengo</b> un gato.",
            "luego (1/1), <b>Tengo</b> un coche.",
        ],
        "children": [],
        "parents": [
            {
                "term": "tener",
                "refs": [
                    "hola (1/1), Ella <b>tiene</b> un perro.",
                    "hola (1/1), No quiero <b>tener</b> nada.",
                ],
            }
        ],
    }, "term tengo"

    tener = repo.find(spanish.id, "tener")
    refs = refsrepo.find_references(tener)
    assert full_refs_to_string(refs) == {
        "term": ["hola (1/1), No quiero <b>tener</b> nada."],
        "children": [
            "hola (1/1), <b>Tengo</b> un gato.",
            "hola (1/1), Ella <b>tiene</b> un perro.",
            "luego (1/1), <b>Tengo</b> un coche.",
        ],
        "parents": [],
    }, "term tener"


@pytest.mark.sentences
def test_multiword_reference(spanish, repo, refsrepo):
    "Ensure zws-delimiters are respected."
    text = make_text("hola", "Yo tengo un gato.", spanish)
    text.read_date = datetime.now()
    db.session.add(text)
    db.session.commit()

    t = Term()
    t.language_id = spanish.id
    t.text = "TENGO UN"
    repo.add(t)
    repo.commit()

    refs = refsrepo.find_references(t)
    assert full_refs_to_string(refs) == {
        "term": ["hola (1/1), Yo <b>tengo un</b> gato."],
        "children": [],
        "parents": [],
    }, "term tengo"


@pytest.mark.sentences
def test_get_references_only_includes_read_texts(spanish, repo, refsrepo):
    "Like it says above."
    text = make_text("hola", "Tengo un gato.  No tengo un perro.", spanish)

    tengo = Term()
    tengo.language_id = spanish.id
    tengo.text = "tengo"
    repo.add(tengo)
    repo.commit()

    refs = refsrepo.find_references(tengo)
    keys = refs.keys()
    for k in keys:
        assert len(refs[k]) == 0, f"{k}, no matches for unread texts"

    text.read_date = datetime.now()
    db.session.add(text)
    db.session.commit()

    refs = refsrepo.find_references(tengo)
    assert len(refs["term"]) == 2, "have refs once text is read"


def _make_read_text(title, body, lang):
    "Make a text, mark it read."
    text = make_text(title, body, lang)
    text.read_date = datetime.now()
    db.session.add(text)
    db.session.commit()
    return text


@pytest.mark.sentences
def test_issue_531_spanish_ref_search_case_insens_normal(spanish, refsrepo):
    "Spanish was finding 'normal' upper/lower chars that sqlite could handle."
    _make_read_text("hola", "TENGO.  tengo.", spanish)
    t = add_terms(spanish, ["tengo"])[0]

    refs = refsrepo.find_references(t)
    assert len(refs["term"]) == 2, "both found"


@pytest.mark.sentences
def test_issue_531_spanish_ref_search_case_insens_accented(spanish, refsrepo):
    "Spanish wasn't finding different case of accented chars."
    _make_read_text("hola", "Ábrelo.  ábrelo.", spanish)
    t = add_terms(spanish, ["ábrelo"])[0]
    refs = refsrepo.find_references(t)
    assert len(refs["term"]) == 2, "both found"


@pytest.mark.sentences
def test_issue_531_turkish_ref_search_is_case_insensitive(turkish, refsrepo):
    "Turkish upper/lower case letters are quite different!."
    _make_read_text("Test", "ışık. Işık", turkish)
    t = add_terms(turkish, ["ışık"])[0]
    refs = refsrepo.find_references(t)
    assert len(refs["term"]) == 2, "both found"


@pytest.mark.sentences
def test_get_references_only_includes_refs_in_same_language(
    spanish, english, repo, refsrepo
):
    "Like it says above."
    text1 = make_text("hola", "Tengo un gato.  No tengo un perro.", spanish)
    text2 = make_text("hola", "Tengo in english.", english)

    tengo = Term()
    tengo.language_id = spanish.id
    tengo.text = "tengo"
    repo.add(tengo)
    repo.commit()

    text1.read_date = datetime.now()
    text2.read_date = datetime.now()
    db.session.add(text1)
    db.session.add(text2)
    db.session.commit()

    refs = refsrepo.find_references(tengo)
    assert len(refs["term"]) == 2, "only have 2 refs (spanish)"
    sentences = [r.sentence for r in refs["term"]]
    expected = [
        "<b>Tengo</b> un gato.",
        "No <b>tengo</b> un perro.",
    ]
    assert sentences == expected


@pytest.mark.sentences
def test_get_references_new_term(spanish, refsrepo):
    "Check references with parents and children."
    text = make_text("hola", "Tengo un gato.", spanish)
    text.read_date = datetime.now()
    db.session.add(text)

    tengo = Term()
    tengo.language_id = spanish.id
    tengo.text = "tengo"

    refs = refsrepo.find_references(tengo)
    assert full_refs_to_string(refs) == {
        "term": ["hola (1/1), <b>Tengo</b> un gato."],
        "children": [],
        "parents": [],
    }, "term tengo"


@pytest.mark.sentences
def test_can_get_references_by_term_id_including_unread(spanish):
    "Like it says above."
    text = make_text("hola", "Tengo un gato.  No tengo un perro.", spanish)
    text.load_sentences()
    db.session.add(text)
    db.session.commit()
    # pylint: disable=unbalanced-tuple-unpacking
    [tengo] = add_terms(spanish, ["tengo"])

    assert_record_count_equals("select * from sentences", 2, "sanity check")

    refsrepo = ReferencesRepository(db.session, limit=1, include_unread=False)
    refs = refsrepo.find_references_by_id(tengo.id)
    sentences = [r.sentence for r in refs["term"]]
    expected = []
    assert sentences == expected, "not including unread texts"

    refsrepo = ReferencesRepository(db.session, limit=1, include_unread=True)
    refs = refsrepo.find_references_by_id(tengo.id)
    sentences = [r.sentence for r in refs["term"]]
    expected = ["<b>Tengo</b> un gato."]
    assert sentences == expected, "including unread"



"""
Anki export.
"""

import json
from flask import (
    Blueprint,
    request,
    jsonify,
    render_template,
    redirect,
    flash,
)
from lute.ankiexport.service import Service
from lute.models.srsexport import SrsExportSpec
from lute.ankiexport.forms import SrsExportSpecForm
from lute.ankiexport.exceptions import AnkiExportConfigurationError
from lute.db import db


bp = Blueprint("ankiexport", __name__, url_prefix="/ankiexport")


@bp.route("/index", methods=["GET", "POST"])
def anki_index():
    "List the exports."
    export_specs = db.session.query(SrsExportSpec).all()
    export_specs_json = [
        {
            "id": spec.id,
            "export_name": spec.export_name,
            "criteria": spec.criteria,
            "deck_name": spec.deck_name,
            "note_type": spec.note_type,
            "field_mapping": spec.field_mapping,
            "active": "yes" if spec.active else "no",
        }
        for spec in export_specs
    ]

    return render_template(
        "/ankiexport/index.html",
        export_specs_json=export_specs_json,
    )


def _handle_form(spec, form_template_name):
    """
    Handle a form post.
    """
    form = SrsExportSpecForm(obj=spec)

    if request.method == "POST":
        anki_settings_json = request.form.get("ankisettings")
        anki_settings = json.loads(anki_settings_json)

        form.anki_deck_names = anki_settings.get("deck_names")
        form.anki_note_types = anki_settings.get("note_types")

        # Have to load the option choices or flask-wtf complains ...
        # ouch.
        form.deck_name.choices = [(f, f) for f in form.anki_deck_names]
        form.note_type.choices = [(f, f) for f in form.anki_note_types.keys()]

    if form.validate_on_submit():
        form.populate_obj(spec)
        db.session.add(spec)
        db.session.commit()
        return redirect("/ankiexport/index", 302)

    return render_template(form_template_name, form=form, spec=spec)


@bp.route("/spec/edit/<int:spec_id>", methods=["GET", "POST"])
def edit_spec(spec_id):
    "Edit a spec."
    spec = db.session.query(SrsExportSpec).filter(SrsExportSpec.id == spec_id).first()
    return _handle_form(spec, "/ankiexport/edit.html")


@bp.route("/spec/new", methods=["GET", "POST"])
def new_spec():
    "Make a new spec."
    spec = SrsExportSpec()
    # Hack ... not sure why this was necessary, given that the model
    # and form both have the default as True.
    if spec.active is None:
        spec.active = True
    return _handle_form(spec, "/ankiexport/new.html")


@bp.route("/spec/delete/<int:spec_id>", methods=["GET", "POST"])
def delete_spec(spec_id):
    "Delete a spec."
    spec = db.session.query(SrsExportSpec).filter(SrsExportSpec.id == spec_id).first()
    db.session.delete(spec)
    db.session.commit()
    flash("Export mapping deleted.")
    return redirect("/ankiexport/index", 302)


@bp.route("/get_card_post_data", methods=["POST"])
def get_ankiconnect_post_data():
    """Get data that the client javascript will post."""
    data = request.get_json()
    word_ids = data["term_ids"]
    termid_sentences = data["termid_sentences"]
    base_url = data["base_url"]
    anki_deck_names = data["deck_names"]
    anki_note_types = data["note_types"]
    export_specs = db.session.query(SrsExportSpec).all()
    svc = Service(anki_deck_names, anki_note_types, export_specs)
    try:
        ret = svc.get_ankiconnect_post_data(
            word_ids, termid_sentences, base_url, db.session
        )
        return jsonify(ret)
    except AnkiExportConfigurationError as ex:
        response = jsonify({"error": str(ex)})
        response.status_code = 400  # Bad Request
        return response


@bp.route("/validate_export_specs", methods=["POST"])
def validate_export_specs():
    """Get data that the client javascript will post."""
    data = request.get_json()
    anki_deck_names = data["deck_names"]
    anki_note_types = data["note_types"]
    export_specs = db.session.query(SrsExportSpec).all()
    svc = Service(anki_deck_names, anki_note_types, export_specs)
    try:
        ret = svc.validate_specs()
        return jsonify(ret)
    except AnkiExportConfigurationError as ex:
        response = jsonify({"error": str(ex)})
        response.status_code = 400  # Bad Request
        return response



"""
Settings routes.
"""

from flask import (
    Blueprint,
    current_app,
    request,
    render_template,
    redirect,
    flash,
    jsonify,
)
from wtforms import BooleanField
from lute.models.language import Language
from lute.models.setting import UserSetting
from lute.models.repositories import UserSettingRepository
from lute.themes.service import Service as ThemeService
from lute.settings.forms import UserSettingsForm, UserShortcutsForm
from lute.settings.current import refresh_global_settings
from lute.settings.hotkey_data import categorized_hotkeys, hotkey_descriptions
from lute.db import db
from lute.parse.mecab_parser import JapaneseParser


bp = Blueprint("settings", __name__, url_prefix="/settings")


@bp.route("/index", methods=["GET", "POST"])
def edit_settings():
    "Edit settings."
    form = UserSettingsForm()

    with current_app.app_context():
        svc = ThemeService(db.session)
        form.current_theme.choices = svc.list_themes()

    ac = current_app.env_config
    if ac.is_docker:
        # User shouldn't change some things with docker.
        kw = {"readonly": True, "style": "background-color: LightGray"}
        # Backup dir gets mounted from host.
        form.backup_dir.render_kw = kw

    repo = UserSettingRepository(db.session)
    if form.validate_on_submit():
        # Update the settings in the database
        for field in form:
            if field.id not in ("csrf_token", "submit"):
                repo.set_value(field.id, field.data)
        db.session.commit()
        refresh_global_settings(db.session)

        flash("Settings updated", "success")
        return redirect("/")

    # Load current settings from the database
    for field in form:
        if field.id != "csrf_token":
            field.data = repo.get_value(field.id)
        if isinstance(field, BooleanField):
            # Hack: set boolean settings to ints, otherwise they're always checked.
            field.data = int(field.data or 0)

    return render_template("settings/form.html", form=form)


@bp.route("/test_mecab", methods=["GET"])
def test_parse():
    """
    Do a test parse for the JapaneseParser using the
    given path string.

    Returns { 'success': tokens }, or { 'error' msg }

    """
    mecab_path = request.args.get("mecab_path", None)
    repo = UserSettingRepository(db.session)
    old_setting = repo.get_value("mecab_path")
    result = {"failure": "tbd"}
    try:
        repo.set_value("mecab_path", mecab_path)
        # Parsing requires a language, even if it's a dummy.
        lang = Language()
        p = JapaneseParser()
        src = "私は元気です"
        toks = p.get_parsed_tokens(src, lang)
        toks = [tok.token for tok in toks if tok.token != "¶"]
        message = f"{src} parsed to [{ ', '.join(toks) }]"
        result = {"result": "success", "message": message}
    except Exception as e:  # pylint: disable=broad-exception-caught
        message = f"{type(e).__name__}: { str(e) }"
        result = {"result": "failure", "message": message}
    finally:
        repo.set_value("mecab_path", old_setting)

    return jsonify(result)


@bp.route("/set/<key>/<value>", methods=["POST"])
def set_key_value(key, value):
    "Set a UserSetting key to value."
    repo = UserSettingRepository(db.session)
    old_value = repo.get_value(key)
    try:
        repo.set_value(key, value)
        result = {"result": "success", "message": "OK"}
    except Exception as e:  # pylint: disable=broad-exception-caught
        message = f"{type(e).__name__}: { str(e) }"
        repo.set_value(key, old_value)
        result = {"result": "failure", "message": message}
    db.session.commit()
    refresh_global_settings(db.session)
    return jsonify(result)


def _get_categorized_hotkeys():
    """
    Return hotkey UserSetting keys and values,
    grouped by category.
    """
    categorized = categorized_hotkeys()
    settings = {h.key: h.value for h in db.session.query(UserSetting).all()}
    return {
        category: {k: settings[k] for k in keylist}
        for category, keylist in categorized.items()
    }


@bp.route("/shortcuts", methods=["GET", "POST"])
def edit_shortcuts():
    "Edit shortcuts."
    repo = UserSettingRepository(db.session)
    form = UserShortcutsForm()
    if form.validate_on_submit():
        # print(request.form, flush=True)
        # Update the settings in the database
        for k, v in request.form.items():
            # print(f"{k} = {v}", flush=True)
            repo.set_value(k, v)
        db.session.commit()
        refresh_global_settings(db.session)
        flash("Shortcuts updated", "success")
        return redirect("/")

    categorized_settings = _get_categorized_hotkeys()
    return render_template(
        "settings/shortcuts.html",
        setting_descs=hotkey_descriptions(),
        categorized_settings=categorized_settings,
    )



"""
book helper routines.
"""

import os
import shutil
from io import StringIO, TextIOWrapper, BytesIO
from datetime import datetime
import uuid
from dataclasses import dataclass
from tempfile import TemporaryFile
import requests
from bs4 import BeautifulSoup
from flask import current_app, flash
from openepub import Epub, EpubError
from pypdf import PdfReader
from subtitle_parser import SrtParser, WebVttParser
from lute.book.model import Repository


class BookImportException(Exception):
    """
    Exception to throw on book import error.
    """

    def __init__(self, message="A custom error occurred", cause=None):
        self.cause = cause
        self.message = message
        super().__init__(message)


@dataclass
class BookDataFromUrl:
    "Data class"
    title: str = None
    source_uri: str = None
    text: str = None


class FileTextExtraction:
    "Utility to extract text from various file formats."

    def get_file_content(self, filename, filestream):
        """
        Get the content of the file.
        """
        _, ext = os.path.splitext(filename)
        ext = (ext or "").lower()

        messages = {
            ".pdf": """
            Note: pdf imports can be inaccurate, due to how PDFs are encoded.
            Please be aware of this while reading.
            """
        }
        msg = messages.get(ext)
        if msg is not None:
            flash(msg, "notice")

        handlers = {
            ".txt": self._get_textfile_content,
            ".epub": self._get_epub_content,
            ".pdf": self._get_pdf_content,
            ".srt": self._get_srt_content,
            ".vtt": self._get_vtt_content,
        }
        handler = handlers.get(ext)
        if handler is None:
            raise ValueError(f'Unknown file extension "{ext}"')
        content = handler(filename, filestream).strip()
        if content == "":
            raise BookImportException(f"{filename} is empty.")
        return content

    def _get_text_stream_content(self, fstream, encoding="utf-8"):
        "Gets content from simple text stream."

        usestream = fstream
        # May have to convert the fstream to a a BytesIO stream.
        # GitHub CI caught this, and per ChatGPT: In Python 3.10,
        # SpooledTemporaryFile no longer automatically gains all
        # file-like methods when rolled over to a regular temporary
        # file. Specifically, it seems that the object lacks the
        # readable method required by TextIOWrapper to validate the
        # stream ...
        #
        # I haven't looked into this deeply, but when running Python
        # 3.10.16 on my mac, "inv accept -k bad_text_files" failed on
        # line "with TextIOWrapper(fstream, encoding=encoding) as
        # decoded:" with "AttributeError: 'SpooledTemporaryFile'
        # object has no attribute 'readable'. Did you mean:
        # 'readline'?"..  Converting usestream to BytesIO fixed it.
        if not hasattr(fstream, "readable"):
            usestream = BytesIO(fstream.read())  # Wrap in BytesIO if needed
        with TextIOWrapper(usestream, encoding=encoding) as decoded:
            return decoded.read()

    def _get_textfile_content(self, filename, filestream):
        "Get content as a single string."
        try:
            return self._get_text_stream_content(filestream)
        except UnicodeDecodeError as e:
            f = filename
            msg = f"{f} is not utf-8 encoding, please convert it to utf-8 first (error: {str(e)})"
            raise BookImportException(message=msg, cause=e) from e

    def _get_epub_content(self, filename, filestream):
        """
        Get the content of the epub as a single string.
        """
        content = ""
        try:
            if hasattr(filestream, "seekable"):
                epub = Epub(stream=filestream)
                content = epub.get_text()
            else:
                # We get a SpooledTemporaryFile from the form but this doesn't
                # implement all file-like methods until python 3.11. So we need
                # to rewrite it into a TemporaryFile
                with TemporaryFile() as tf:
                    filestream.seek(0)
                    tf.write(filestream.read())
                    epub = Epub(stream=tf)
                    content = epub.get_text()
        except EpubError as e:
            msg = f"Could not parse {filename} (error: {str(e)})"
            raise BookImportException(message=msg, cause=e) from e
        return content

    def _get_pdf_content(self, filename, filestream):
        "Get content as a single string from a PDF file using PyPDF2."
        content = ""
        try:
            pdf_reader = PdfReader(filestream)
            for page in pdf_reader.pages:
                content += page.extract_text()
            return content
        except Exception as e:
            msg = f"Could not parse {filename} (error: {str(e)})"
            raise BookImportException(message=msg, cause=e) from e

    def _get_srt_content(self, filename, filestream):
        """
        Get the content of the srt as a single string.
        """
        content = ""
        try:
            srt_content = self._get_text_stream_content(filestream, "utf-8-sig")
            parser = SrtParser(StringIO(srt_content))
            parser.parse()
            content = "\n".join(subtitle.text for subtitle in parser.subtitles)
            return content
        except Exception as e:
            msg = f"Could not parse {filename} (error: {str(e)})"
            raise BookImportException(message=msg, cause=e) from e

    def _get_vtt_content(self, filename, filestream):
        """
        Get the content of the vtt as a single string.
        """
        content = ""
        try:
            vtt_content = self._get_text_stream_content(filestream, "utf-8-sig")
            # Check if it is from YouTube
            lines = vtt_content.split("\n")
            if lines[1].startswith("Kind:") and lines[2].startswith("Language:"):
                vtt_content = "\n".join(lines[:1] + lines[3:])
            parser = WebVttParser(StringIO(vtt_content))
            parser.parse()
            content = "\n".join(subtitle.text for subtitle in parser.subtitles)
            return content
        except Exception as e:
            msg = f"Could not parse {filename} (error: {str(e)})"
            raise BookImportException(message=msg, cause=e) from e


class Service:
    "Service."

    def _unique_fname(self, filename):
        """
        Return secure name pre-pended with datetime string.
        """
        current_datetime = datetime.now()
        formatted_datetime = current_datetime.strftime("%Y%m%d_%H%M%S")
        _, ext = os.path.splitext(filename)
        ext = (ext or "").lower()
        newfilename = uuid.uuid4().hex
        return f"{formatted_datetime}_{newfilename}{ext}"

    def save_audio_file(self, audio_file_field_data):
        """
        Save the file to disk, return its filename.
        """
        filename = self._unique_fname(audio_file_field_data.filename)
        fp = os.path.join(current_app.env_config.useraudiopath, filename)
        audio_file_field_data.save(fp)
        return filename

    def book_data_from_url(self, url):
        """
        Parse the url and load source data for a new Book.
        This returns a domain object, as the book is still unparsed.
        """
        s = None
        try:
            timeout = 20  # seconds
            response = requests.get(url, timeout=timeout)
            response.raise_for_status()
            s = response.text
        except requests.exceptions.RequestException as e:
            msg = f"Could not parse {url} (error: {str(e)})"
            raise BookImportException(message=msg, cause=e) from e

        soup = BeautifulSoup(s, "html.parser")
        extracted_text = []

        # Add elements in order found.
        for element in soup.descendants:
            if element.name in ("h1", "h2", "h3", "h4", "p"):
                extracted_text.append(element.text)

        title_node = soup.find("title")
        orig_title = title_node.string if title_node else url

        short_title = orig_title[:150]
        if len(orig_title) > 150:
            short_title += " ..."

        b = BookDataFromUrl()
        b.title = short_title
        b.source_uri = url
        b.text = "\n\n".join(extracted_text)
        return b

    def import_book(self, book, session):
        """
        Save the book as a dbbook, parsing and saving files as needed.
        Returns new book created.
        """

        def _raise_if_file_missing(p, fldname):
            if not os.path.exists(p):
                raise BookImportException(f"Missing file {p} given in {fldname}")

        def _raise_if_none(p, fldname):
            if p is None:
                raise BookImportException(f"Must set {fldname}")

        fte = FileTextExtraction()
        if book.text_source_path:
            _raise_if_file_missing(book.text_source_path, "text_source_path")
            tsp = book.text_source_path
            with open(tsp, mode="rb") as stream:
                book.text = fte.get_file_content(tsp, stream)

        if book.text_stream:
            _raise_if_none(book.text_stream_filename, "text_stream_filename")
            book.text = fte.get_file_content(
                book.text_stream_filename, book.text_stream
            )

        if book.audio_source_path:
            _raise_if_file_missing(book.audio_source_path, "audio_source_path")
            newname = self._unique_fname(book.audio_source_path)
            fp = os.path.join(current_app.env_config.useraudiopath, newname)
            shutil.copy(book.audio_source_path, fp)
            book.audio_filename = newname

        if book.audio_stream:
            _raise_if_none(book.audio_stream_filename, "audio_stream_filename")
            newname = self._unique_fname(book.audio_stream_filename)
            fp = os.path.join(current_app.env_config.useraudiopath, newname)
            with open(fp, mode="wb") as fcopy:  # Use "wb" to write in binary mode
                while chunk := book.audio_stream.read(
                    8192
                ):  # Read the stream in chunks (e.g., 8 KB)
                    fcopy.write(chunk)
            book.audio_filename = newname

        repo = Repository(session)
        dbbook = repo.add(book)
        repo.commit()
        return dbbook



"""
Srs export entity.
"""

from lute.db import db


class SrsExportSpec(db.Model):
    """
    Srs export spec entity.
    """

    __tablename__ = "srsexportspecs"

    id = db.Column("SrsID", db.Integer, primary_key=True)
    export_name = db.Column(
        "SrsExportName", db.String(200), nullable=False, unique=True
    )
    criteria = db.Column("SrsCriteria", db.String(1000), nullable=False)
    deck_name = db.Column("SrsDeckName", db.String(200), nullable=False)
    note_type = db.Column("SrsNoteType", db.String(200), nullable=False)
    field_mapping = db.Column("SrsFieldMapping", db.String(1000), nullable=False)
    active = db.Column("SrsActive", db.Boolean, nullable=False, default=True)



"""
Criteria.  Given a string with selection criteria, evaluates it
with a term, returning True or False.
"""

from typing import Callable, Iterable
import pyparsing as pp
from pyparsing import (
    infixNotation,
    opAssoc,
    Keyword,
    Word,
    # ParserElement,
    nums,
    one_of,
    quotedString,
    QuotedString,
    Suppress,
    Literal,
)
from pyparsing.exceptions import ParseException
from lute.models.term import Term
from lute.models.language import Language
from lute.ankiexport.exceptions import AnkiExportConfigurationError


def evaluate_criteria(s, term):
    "Parse the criteria, return True or False for the given term."
    # pylint: disable=too-many-locals

    if (s or "").strip() == "":
        return True

    def has_any_matching_tags(tagvals):
        term_tags = [t.text for t in term.term_tags]
        return any(e in term_tags for e in tagvals)

    def has_any_matching_parent_tags(tagvals):
        ptags = []
        for p in term.parents:
            ptags.extend([t.text for t in p.term_tags])
        return any(e in ptags for e in tagvals)

    def has_any_matching_all_tags(tagvals):
        alltags = [t.text for t in term.term_tags]
        for p in term.parents:
            alltags.extend([t.text for t in p.term_tags])
        return any(e in alltags for e in tagvals)

    def matches_lang(lang):
        return term.language.name == lang[0]

    def check_has_images():
        "True if term or any parent has image."
        pi = [p.get_current_image() is not None for p in term.parents]
        return term.get_current_image() is not None or any(pi)

    def check_has(args):
        "Check has:x"
        has_item = args[0]
        if has_item == "image":
            return check_has_images()
        raise RuntimeError(f"Unhandled has check for {has_item}")

    def get_binary_operator(opstring):
        "Return lambda matching op."
        opMap = {
            "<": lambda a, b: a < b,
            "<=": lambda a, b: a <= b,
            ">": lambda a, b: a > b,
            ">=": lambda a, b: a >= b,
            "!=": lambda a, b: a != b,
            "=": lambda a, b: a == b,
            "==": lambda a, b: a == b,
        }
        return opMap[opstring]

    def check_parent_count(args):
        "Check parents."
        opstring, val = args
        oplambda = get_binary_operator(opstring)
        pcount = len(term.parents)
        return oplambda(pcount, val)

    def check_status_val(args):
        "Check status."
        opstring, val = args
        oplambda = get_binary_operator(opstring)
        return oplambda(term.status, val)

    ### class BoolNot:
    ###     "Not unary operator."
    ###     def __init__(self, t):
    ###         self.arg = t[0][1]
    ###     def __bool__(self) -> bool:
    ###         v = bool(self.arg)
    ###         return not v
    ###     def __str__(self) -> str:
    ###         return "~" + str(self.arg)
    ###     __repr__ = __str__

    class BoolBinOp:
        "Binary operation."
        repr_symbol: str = ""
        eval_fn: Callable[[Iterable[bool]], bool] = lambda _: False

        def __init__(self, t):
            self.args = t[0][0::2]

        def __str__(self) -> str:
            sep = f" {self.repr_symbol} "
            return f"({sep.join(map(str, self.args))})"

        def __bool__(self) -> bool:
            return self.eval_fn(bool(a) for a in self.args)

    class BoolAnd(BoolBinOp):
        repr_symbol = "&"
        eval_fn = all

    class BoolOr(BoolBinOp):
        repr_symbol = "|"
        eval_fn = any

    quoteval = QuotedString(quoteChar='"')
    quotedString.setParseAction(pp.removeQuotes)
    list_of_values = pp.delimitedList(quotedString)

    tagvallist = Suppress("[") + list_of_values + Suppress("]")
    tagcrit = tagvallist | quoteval

    tag_matcher = Suppress(Literal("tags") + Literal(":")) + tagcrit
    parents_tag_matcher = Suppress(Literal("parents.tags") + Literal(":")) + tagcrit
    all_tag_matcher = Suppress(Literal("all.tags") + Literal(":")) + tagcrit

    lang_matcher = Suppress("language") + Suppress(":") + quoteval

    has_options = Literal("image")
    has_matcher = Suppress("has") + Suppress(":") + has_options

    comparison_op = one_of("< <= > >= != = == <>")
    integer = Word(nums).setParseAction(lambda x: int(x[0]))

    parent_count_matcher = (
        Suppress("parents")
        + Suppress(".")
        + Suppress("count")
        + comparison_op
        + integer
    )

    status_matcher = Suppress("status") + comparison_op + integer

    and_keyword = Keyword("and")
    or_keyword = Keyword("or")

    multi_check = infixNotation(
        tag_matcher.set_parse_action(has_any_matching_tags)
        | parents_tag_matcher.set_parse_action(has_any_matching_parent_tags)
        | all_tag_matcher.set_parse_action(has_any_matching_all_tags)
        | lang_matcher.set_parse_action(matches_lang)
        | has_matcher.set_parse_action(check_has)
        | parent_count_matcher.set_parse_action(check_parent_count)
        | status_matcher.set_parse_action(check_status_val),
        [
            (and_keyword, 2, opAssoc.LEFT, BoolAnd),
            (or_keyword, 2, opAssoc.LEFT, BoolOr),
        ],
    )

    try:
        result = multi_check.parseString(s, parseAll=True)
        return bool(result[0])
    except pp.ParseException as ex:
        msg = f"Criteria syntax error at position {ex.loc} or later: {ex.line}"
        raise AnkiExportConfigurationError(msg) from ex


def validate_criteria(criteria):
    "Check criteria with a dummy Term."
    term = Term(Language(), "")
    try:
        evaluate_criteria(criteria, term)
    except ParseException as ex:
        msg = f'Invalid criteria "{ex.line}"'
        raise AnkiExportConfigurationError(msg) from ex



"""
Service tests.
"""

import json
from unittest.mock import Mock
import pytest
from lute.models.srsexport import SrsExportSpec
from lute.ankiexport.service import Service

# pylint: disable=missing-function-docstring


@pytest.fixture(name="export_spec")
def fixture_spec():
    spec = SrsExportSpec()
    spec.id = 1
    spec.export_name = "export_name"
    spec.criteria = 'language:"German"'
    spec.deck_name = "good_deck"
    spec.note_type = "good_note"
    spec.field_mapping = json.dumps({"a": "{ language }"})
    spec.active = True
    return spec


def test_validate_returns_empty_hash_if_all_ok(export_spec):
    anki_decks = ["good_deck"]
    anki_notes = {"good_note": ["a", "b"]}
    svc = Service(anki_decks, anki_notes, [export_spec])
    result = svc.validate_specs()
    assert len(result) == 0, "No problems"
    msg = svc.validate_specs_failure_message()
    assert len(msg) == 0, "failure msg"


@pytest.mark.parametrize(
    "prop_name,prop_value,expected_error",
    [
        (
            "criteria",
            'lanxxguage:"German"',
            'Criteria syntax error at position 0 or later: lanxxguage:"German"',
        ),
        ("deck_name", "missing_deck", 'No deck name "missing_deck"'),
        ("note_type", "missing_note", 'No note type "missing_note"'),
        (
            "field_mapping",
            json.dumps({"xx": "{ language }"}),
            "Note type good_note does not have field(s): xx",
        ),
        (
            "field_mapping",
            json.dumps({"a": "{ bad_value }"}),
            'Invalid field mapping "bad_value"',
        ),
        (
            "field_mapping",
            "this_is_not_valid_json",
            "Mapping is not valid json",
        ),
    ],
)
def test_validate_spec_returns_array_of_errors(
    prop_name, prop_value, expected_error, export_spec
):
    anki_decks = ["good_deck"]
    anki_notes = {"good_note": ["a", "b"]}
    setattr(export_spec, prop_name, prop_value)
    svc = Service(anki_decks, anki_notes, [export_spec])
    result = svc.validate_spec(export_spec)
    assert result == [expected_error]

    export_spec.active = False
    assert len(svc.validate_spec(export_spec)) == 0, "no errors for inactive spec"


def test_validate_specs_returns_dict_of_export_ids_and_errors(export_spec):
    anki_decks = ["good_deck"]
    anki_notes = {"good_note": ["a", "b"]}
    export_spec.deck_name = "missing_deck"
    svc = Service(anki_decks, anki_notes, [export_spec])
    result = svc.validate_specs()
    assert result == {export_spec.id: 'No deck name "missing_deck"'}

    msg = svc.validate_specs_failure_message()
    assert msg == ['export_name: No deck name "missing_deck"'], "failure msg"


def test_validate_only_checks_active_specs(export_spec):
    anki_decks = ["good_deck"]
    anki_notes = {"good_note": ["a", "b"]}
    export_spec.criteria = "xxx={yyy}"
    svc = Service(anki_decks, anki_notes, [export_spec])
    result = svc.validate_specs()
    assert export_spec.id in result, "should have a problem, sanity check"

    export_spec.active = False
    result = svc.validate_specs()
    assert len(result) == 0, "No problems"
    msg = svc.validate_specs_failure_message()
    assert len(msg) == 0, "failure msg"


@pytest.fixture(name="term")
def fixture_term():
    zws = "\u200B"
    term = Mock()
    term.id = 1
    term.text = f"test{zws} {zws}term"
    term.language.name = "German"
    term.language.id = 42
    term.get_current_image.return_value = "image.jpg"
    term.parents = []
    term.term_tags = [Mock(text="noun"), Mock(text="verb")]
    term.translation = f"example{zws} {zws}translation"
    return term


def test_smoke_ankiconnect_post_data_for_term(term, export_spec):
    anki_decks = ["good_deck"]
    anki_notes = {"good_note": ["a", "b", "c", "d"]}
    export_spec.field_mapping = json.dumps(
        {
            "a": "{ language }",
            "b": "{ image }",
            "c": "{ term }",
            "d": "{ sentence }",
        }
    )
    svc = Service(anki_decks, anki_notes, [export_spec])
    result = svc.validate_specs()
    assert len(result) == 0, "No problems, sanity check"

    sentence_lookup = Mock()
    sentence_lookup.get_sentence_for_term.return_value = "Example sentence."

    pd = svc.get_ankiconnect_post_data_for_term(term, "http://x:42", sentence_lookup)
    assert len(pd) != 0, "Got some post data"

    expected = {
        "export_name": {
            "action": "multi",
            "params": {
                "actions": [
                    {
                        "action": "storeMediaFile",
                        "params": {
                            "filename": "LUTE_TERM_1.jpg",
                            "url": "http://x:42/userimages/42/image.jpg",
                        },
                    },
                    {
                        "action": "addNote",
                        "params": {
                            "note": {
                                "deckName": "good_deck",
                                "modelName": "good_note",
                                "fields": {
                                    "a": "German",
                                    "b": '<img src="LUTE_TERM_1.jpg">',
                                    "c": "test term",
                                    "d": "Example sentence.",
                                },
                                "tags": ["lute", "noun", "verb"],
                            }
                        },
                    },
                ]
            },
        }
    }

    # print("actual")
    # print(pd)
    # print("expected")
    # print(expected)
    assert pd == expected, "PHEW!"


def test_smoke_ankiconnect_post_data_for_term_without_image(term, export_spec):
    term.get_current_image.return_value = None

    anki_decks = ["good_deck"]
    anki_notes = {"good_note": ["a", "b", "c", "d"]}
    export_spec.field_mapping = json.dumps(
        {
            "a": "{ language }",
            "b": "{ image }",
            "c": "{ term }",
            "d": "{ sentence }",
        }
    )
    svc = Service(anki_decks, anki_notes, [export_spec])
    result = svc.validate_specs()
    assert len(result) == 0, "No problems, sanity check"

    sentence_lookup = Mock()
    sentence_lookup.get_sentence_for_term.return_value = "Example sentence."

    pd = svc.get_ankiconnect_post_data_for_term(term, "http://x:42", sentence_lookup)
    assert len(pd) != 0, "Got some post data"

    expected = {
        "export_name": {
            "action": "multi",
            "params": {
                "actions": [
                    {
                        "action": "addNote",
                        "params": {
                            "note": {
                                "deckName": "good_deck",
                                "modelName": "good_note",
                                "fields": {
                                    "a": "German",
                                    # "b": "", image not posted at all.
                                    "c": "test term",
                                    "d": "Example sentence.",
                                },
                                "tags": ["lute", "noun", "verb"],
                            }
                        },
                    },
                ]
            },
        }
    }

    # print("actual")
    # print(pd)
    # print("expected")
    # print(expected)
    assert pd == expected, "PHEW!"
